from girder import events, plugin
from girder.constants import AccessType
from girder.models.file import File
from girder.models.item import Item
from girder.models.folder import Folder
from .api.app import App
from . import constants, utilities
from .models.package import Package as PackageModel

from girder_hashsum_download import SUPPORTED_ALGORITHMS

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


def _onDownloadFileComplete(event):
    """
    Increment download count associated with item revision associated with the event.

    Count is stored in the ``downloadStats`` metadata organized as a json document
    set in the parent release folder.

    See :func:`utilities.getReleaseFolder()`.
    """
    item = Item().load(event.info['file']['itemId'], force=True)

    if not utilities.isSlicerPackages(item):
        return

    release = utilities.getReleaseFolder(item)
    if release is None:
        return

    meta = item['meta']
    is_draft_release = release['name'] == constants.DRAFT_RELEASE_NAME
    is_extension_item = 'app_revision' in meta

    if is_extension_item:
        folder_name = constants.EXTENSIONS_FOLDER_NAME
        if is_draft_release:
            field_template = 'meta.downloadStats.{app_revision}.{folder_name}.{baseName}.{os}.{arch}'
        else:
            field_template = 'meta.downloadStats.{folder_name}.{baseName}.{os}.{arch}'
    else:
        folder_name = 'applications'
        if is_draft_release:
            field_template = 'meta.downloadStats.{revision}.{folder_name}.{os}.{arch}'
        else:
            field_template = 'meta.downloadStats.{folder_name}.{os}.{arch}'

    Folder().increment(
        query={'_id': release['_id']},
        field=field_template.format(folder_name=folder_name, **meta),
        amount=1)


def _onItemSavedOrCopied(event):
    """
    Set or update "release" metadata when an application package item is
    moved or copied into or out of a release folder.

    See :func:`utilities.isSlicerPackages()` and :func:`utilities.getReleaseFolder()`.
    """
    item = Item().load(event.info['_id'], force=True)

    if not utilities.isSlicerPackages(item):
        return

    release = utilities.getReleaseFolder(item, force=True)
    if release is None:
        return

    meta = item['meta']

    is_draft_release = release['name'] == constants.DRAFT_RELEASE_NAME
    is_extension_item = 'app_revision' in meta

    if is_extension_item:
        return

    if is_draft_release:
        if 'release' not in meta:
            return
        del meta['release']

    else:
        if meta.get('release') == release['name']:
            return
        meta['release'] = release['name']

    PackageModel().setMetadata(item, meta)


def _onFileEvent(event):
    """Update checksum metadata for an application or extension package item when a file is saved, copied
    or about to be removed.

    See :func:`utilities.isChildOfSlicerPackages()`.
    """
    file = File().load(event.info['_id'], force=True)
    if not utilities.isChildOfSlicerPackages(file):
        return

    item = Item().load(file['itemId'], force=True)

    # Both application and extension packages are expected to have these metadata
    if not all(meta in item.get('meta', {}) for meta in ['app_id', 'os', 'arch', 'revision']):
        return

    if event.name == "model.file.save.after":
        if Item().childFiles(item).count() > 1:
            # If the update is not related to the first file, ignore
            return

        # Collect checksums associated if the file
        checksums = {algo: file[algo] for algo in SUPPORTED_ALGORITHMS if algo in file}

    if event.name == "model.file.remove":
        item_files = list(Item().childFiles(item))

        if len(item_files) > 2:
            return

        elif len(item_files) == 2:
            # If after removing this file, there are only one file left, update the item with
            # the checksum of the remaining file.
            remaining_file = [_file for _file in item_files if _file["_id"] != file["_id"]][0]

            # Collect checksums associated if the remaining file
            checksums = {algo: remaining_file[algo] for algo in SUPPORTED_ALGORITHMS if algo in remaining_file}

        elif len(item_files) == 1:
            # If after removing this file, there are no file left, set the checksum to empty string.
            # Collect checksums associated if the remaining file
            checksums = {algo: "" for algo in SUPPORTED_ALGORITHMS}

    # Update metadata overwriting existing checksum values if any
    meta = {**item['meta'], **checksums}

    Item().setMetadata(item, meta)


def _onReleaseFolderNameUpdated(event):
    """
    Update "release" metadata on all application package items in a release folder when its name is changed.

    See :func:`utilities.isReleaseFolder()`.
    """
    folder = Folder().load(event.info['_id'], force=True)
    if not utilities.isReleaseFolder(folder):
        return

    release = folder
    if release['name'] == constants.DRAFT_RELEASE_NAME:
        return

    filters = {
        '$and': [
            {'meta.app_id': {'$exists': True}},
            {'meta.os': {'$exists': True}},
            {'meta.arch': {'$exists': True}},
            {'meta.revision': {'$exists': True}}],
    }
    items = Folder().childItems(
        folder=release,
        filters=filters,
        level=AccessType.READ,
    )
    for item in items:
        item_meta = item['meta']
        if item_meta.get('release') == release['name']:
            continue
        item_meta['release'] = release['name']
        PackageModel().setMetadata(item, item_meta)


class GirderPlugin(plugin.GirderPlugin):
    DISPLAY_NAME = 'Slicer Package Manager'

    def load(self, info):
        # add plugin loading logic here
        plugin.getPlugin('hashsum_download').load(info)

        info['apiRoot'].app = App()
        info['serverRoot'].updateHtmlVars({'title': 'Slicer package manager'})

        # Download statistics
        events.bind('model.file.download.complete', 'slicer_package_manager', _onDownloadFileComplete)

        # Update item "release" metadata
        events.bind('model.item.save.after', 'slicer_package_manager', _onItemSavedOrCopied)
        events.bind('model.item.copy.after', 'slicer_package_manager', _onItemSavedOrCopied)
        events.bind('model.folder.save.after', 'slicer_package_manager', _onReleaseFolderNameUpdated)

        # Update item metadata with file checksums
        events.bind('model.file.save.after', 'slicer_package_manager', _onFileEvent)
        events.bind('model.file.remove', 'slicer_package_manager', _onFileEvent)

        # Mongo indexes
        Item().ensureIndex('meta.baseName')
        Item().ensureIndex('meta.os')
        Item().ensureIndex('meta.arch')
        Item().ensureIndex('meta.app_revision')
        Item().ensureIndex('updated')
        Folder().ensureIndex('meta.downloadExtensions')



import os

from shutil import copyfile

from slicer_package_manager.constants import (
    APPLICATION_PACKAGE_TEMPLATE_NAME,
    EXTENSION_PACKAGE_TEMPLATE_NAME,
)


FIXTURE_DIR = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'data',
    )

APPS = ['application']

RELEASES = [
    {
        'app_name': APPS[0],
        'name': 'release1',
        'revision': '0005',
        'version': '0.1.0',
    },
]

DRAFT_RELEASES = [
    {
        'revision': '0000',
        'version': '0.2.0',
    },
    {
        'revision': '0001',
        'version': '0.3.0',
    },
]

RELEASE_EXTENSIONS = [
    {
        'filepath': 'extension0.tar.gz',
        'meta': {
            'os': 'linux',
            'arch': 'i386',
            'baseName': 'Ext0',
            'repository_type': 'git',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '35333',
            'app_revision': RELEASES[0]['revision'],
            'description': 'Extension for Slicer 4',
        },
    },
]

DRAFT_EXTENSIONS = [
    {
        'filepath': 'extension1.tar.gz',
        'meta': {
            'os': 'win',
            'arch': 'i386',
            'baseName': 'Ext1',
            'repository_type': 'git',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '54342',
            'app_revision': DRAFT_RELEASES[0]['revision'],
            'description': 'Extension for Slicer 4 new version',
        },
    },
    {
        'filepath': 'extension2.tar.gz',
        'meta': {
            'os': 'linux',
            'arch': 'amd64',
            'baseName': 'Ext2',
            'repository_type': 'gitlab',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '542',
            'app_revision': DRAFT_RELEASES[1]['revision'],
            'description': 'Extension for Slicer 4 new version',
        },
    },
    {
        'filepath': 'extension3.tar.gz',
        'meta': {
            'os': 'macosx',
            'arch': 'amd64',
            'baseName': 'Ext2',
            'repository_type': 'gitlab',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '542',
            'app_revision': DRAFT_RELEASES[1]['revision'],
            'description': 'Extension for Slicer 4 new version',
        },
    },
    {
        'filepath': 'extension4.tar.gz',
        'meta': {
            'os': 'macosx',
            'arch': 'i386',
            'baseName': 'Ext2',
            'repository_type': 'gitlab',
            'repository_url': 'http://slicer.com/extension/Ext',
            'revision': '542',
            'app_revision': DRAFT_RELEASES[1]['revision'],
            'description': 'Extension for Slicer 4 new version',
        },
    },
]

EXTENSIONS = []
EXTENSIONS.extend(RELEASE_EXTENSIONS)
EXTENSIONS.extend(DRAFT_EXTENSIONS)

for extension in EXTENSIONS:
    extension['name'] = EXTENSION_PACKAGE_TEMPLATE_NAME.format(**extension['meta'])


RELEASE_PACKAGES = [
    {
        'filepath': 'pkg0.dmg',
        'meta': {
            'os': 'macosx',
            'arch': 'amd64',
            'baseName': 'pkg0',
            'repository_type': 'gitlab',
            'repository_url': 'https://slicer4.com',
            'revision': RELEASES[0]['revision'],
            'version': RELEASES[0]['version'],
        },
    },
]

DRAFT_PACKAGES = [
    {
        'filepath': 'pkg1.exe',
        'meta': {
            'os': 'win',
            'arch': 'i386',
            'baseName': 'pkg1',
            'repository_type': 'gitlab',
            'repository_url': 'https://slicer4.com',
            'revision': DRAFT_RELEASES[0]['revision'],
            'version': DRAFT_RELEASES[0]['version'],
        },
    },
    {
        'filepath': 'pkg2.tar.gz',
        'meta': {
            'os': 'linux',
            'arch': 'amd64',
            'baseName': 'pkg2',
            'repository_type': 'git',
            'repository_url': 'git://slicer4.com',
            'revision': DRAFT_RELEASES[0]['revision'],
            'version': DRAFT_RELEASES[0]['version'],
        },
    },
]

PACKAGES = []
PACKAGES.extend(RELEASE_PACKAGES)
PACKAGES.extend(DRAFT_PACKAGES)

for package in PACKAGES:
    package['name'] = APPLICATION_PACKAGE_TEMPLATE_NAME.format(**package['meta'])

expectedDownloadStats = {
    DRAFT_RELEASES[0]['revision']: {
        'applications': {
            'win': {
                'i386': 1,
            },
            'linux': {
                'amd64': 1,
            },
        },
        'extensions': {
            'Ext1': {
                'win': {
                    'i386': 1,
                },
            },
        },
    },
    DRAFT_RELEASES[1]['revision']: {
        'extensions': {
            'Ext2': {
                'linux': {
                    'amd64': 1,
                },
                'macosx': {
                    'amd64': 1,
                    'i386': 1,
                },
            },
        },
    },
    RELEASES[0]['revision']: {
        'applications': {
            'macosx': {
                'amd64': 1,
            },
        },
        'extensions': {
            'Ext0': {
                'linux': {
                    'i386': 1,
                },
            },
        },
    },
}


def computeContentChecksum(algo, content):
    """Compute digest of ``content`` using ``algo``.

    Supported hashing algorithms are SHA256, SHA512, and MD5.

    :raises ValueError: if algo is unknown.
    """
    import hashlib

    if algo not in ['SHA256', 'SHA512', 'MD5']:
        msg = f"unsupported hashing algorithm {algo}"
        raise ValueError(msg)

    digest = hashlib.new(algo)
    digest.update(content)
    return digest.hexdigest()


def computeFileChecksum(algo, filePath):
    """Compute digest of ``filePath`` using ``algo``.

    Supported hashing algorithms are SHA256, SHA512, and MD5.
    It internally reads the file by chunk of 8192 bytes.

    :raises ValueError: if algo is unknown.
    :raises IOError: if filePath does not exist.
    """
    import hashlib

    if algo not in ['SHA256', 'SHA512', 'MD5']:
        msg = f"unsupported hashing algorithm {algo}"
        raise ValueError(msg)

    with open(filePath, 'rb') as content:
        digest = hashlib.new(algo)
        while True:
            chunk = content.read(8192)
            if not chunk:
                break
            digest.update(chunk)
        return digest.hexdigest()


class ExternalData:

    def __init__(self, objectStorePath):
        self._downloadPercent = 0
        self._objectStorePath = objectStorePath

    @staticmethod
    def _humanFormatSize(size):
        """Convert size in bytes to human readable string including unit suffix.

        Adapted from
        https://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
        """
        for x in ['bytes', 'KB', 'MB', 'GB']:
            if -1024.0 < size < 1024.0:
                return "%3.1f %s" % (size, x)
            size /= 1024.0
        return "%3.1f %s" % (size, 'TB')

    @staticmethod
    def _extractAlgoAndDigest(checksum):
        """Given a checksum string formatted as ``<algo>:<digest>`` returns the tuple ``(algo, digest)``.

        ``<algo>`` is expected to be `SHA256`, `SHA512`, or `MD5`.
        ``<digest>`` is expected to be the full length hexadecimal digest.

        :raises ValueError: if checksum is incorrectly formatted.
        """
        if checksum is None:
            return None, None
        if len(checksum.split(':')) != 2:
            msg = f"invalid checksum '{checksum}'. Expected format is '<algo>:<digest>'."
            raise ValueError(msg)
        (algo, digest) = checksum.split(':')
        expected_algos = ['SHA256', 'SHA512', 'MD5']
        if algo not in expected_algos:
            msg = f"Invalid algo '{algo}'. Algo must be one of {', '.join(expected_algos)}."
            raise ValueError(msg)
        expected_digest_length = {'SHA256': 64, 'SHA512': 128, 'MD5': 32}
        if len(digest) != expected_digest_length[algo]:
            msg  = (
                f"Invalid digest length {len(digest)}. "
                f"Expected digest length for {algo} is {expected_digest_length[algo]}."
            )
            raise ValueError(msg)
        return algo, digest

    def _reportHook(self, blocksSoFar, blockSize, totalSize):
        # we clamp to 100% because the blockSize might be larger than the file itself
        percent = min(int((100. * blocksSoFar * blockSize) / totalSize), 100)
        if percent == 100 or (percent - self._downloadPercent >= 10):
            # we clamp to totalSize when blockSize is larger than totalSize
            humanSizeSoFar = self._humanFormatSize(min(blocksSoFar * blockSize, totalSize))
            humanSizeTotal = self._humanFormatSize(totalSize)
            print('Downloaded %s (%d%% of %s)...' % (humanSizeSoFar, percent, humanSizeTotal))
            self._downloadPercent = percent

    def _downloadFile(self, uri, name, checksum):
        """
        :param uri: Download URL.
        :param name: File name that will be downloaded.
        :param checksum: Checksum formatted as ``<algo>:<digest>`` to verify the downloaded file.
            For example, ``SHA256:cc211f0dfd9a05ca3841ce1141b292898b2dd2d3f08286affadf823a7e58df93``.
        """
        self._downloadPercent = 0
        filePath = self._objectStorePath + '/' + name
        (algo, digest) = self._extractAlgoAndDigest(checksum)
        if not os.path.exists(filePath) or os.stat(filePath).st_size == 0:
            import urllib.error
            import urllib.parse
            import urllib.request
            print('Requesting download %s from %s ...' % (name, uri))
            try:
                urllib.request.urlretrieve(uri, filePath, self._reportHook)
                print('Download finished')
            except OSError as exc:
                msg = f"Failed to download {uri} to {filePath}: {exc}"
                raise ValueError(msg)

            if algo is not None:
                print('Verifying checksum')
                current_digest = computeFileChecksum(algo, filePath)
                if current_digest != digest:
                    print(
                        'Checksum verification failed. Computed checksum %s different from expected checksum %s' % (
                            current_digest, digest))
                    os.remove(filePath)
                else:
                    self._downloadPercent = 100
                    print('Checksum OK')
        else:
            if algo is not None:
                print('Verifying checksum')
                current_digest = computeFileChecksum(algo, filePath)
                if current_digest != digest:
                    print('File already exists in cache but checksum is different - re-downloading it.')
                    os.remove(filePath)
                    return self._downloadFile(uri, name, checksum)
                else:
                    self._downloadPercent = 100
                    print('File already exists and checksum is OK - reusing it.')
            else:
                self._downloadPercent = 100
                print('File already exists in cache - reusing it.')
        return filePath

    def download(self, uri, fileName, checksum):
        """Download data into storage directory.

        :param uri: Download URL.
        :param fileName: File name that will be downloaded.
        :param checksum: Checksum formatted as ``<algo>:<digest>`` to verify the downloaded file.
            For example, ``SHA256:cc211f0dfd9a05ca3841ce1141b292898b2dd2d3f08286affadf823a7e58df93``.
        """
        if not os.access(self._objectStorePath, os.W_OK):
            os.makedirs(self._objectStorePath, exist_ok=True)
            assert os.access(self._objectStorePath, os.W_OK)
        maximumAttemptsCount = 3
        errors = set()
        for _attemptsCount in range(maximumAttemptsCount):
            try:
                return self._downloadFile(uri, fileName, checksum)
            except ValueError as exc:
                errors.add(str(exc))
                continue
        msg = (
            f"Download of {fileName} failed for {maximumAttemptsCount} attempts\n"
            f"  uri: {uri}\n"
            f"  errors: {', '.join(map(str, errors))}"
        )
        raise RuntimeError(msg)


def downloadExternals(key_files, dest_dir):
    """
    Download the data files identified by the key files from https://data.kitware.com.

    :param key_files: List of the data file with a “.sha512” extension appended to the file name.
    :param dest_dir: Directory where files are downloaded and stored
    :return: dest_dir
    """
    # Collect externals in map where keys are "fileName" and values are "(<algo>, <checksum>)"
    externals = {}
    for key_file in key_files:
        ext_not_found = []
        for ext in ['sha512']:
            external = key_file + "." + ext
            if os.path.exists(external):
                break
            else:
                ext_not_found.append(ext)
                external = None
        if external is None:
            msg = f"{key_file}.{', '.join(ext_not_found)} not found"
            raise ValueError(msg)
        with open(external) as content:
            hashsum = content.read().strip()
        externals[os.path.basename(key_file)] = (ext, hashsum)

    external_data_dir = os.environ.get(
        'GIRDER_TEST_DATA_PREFIX', os.path.join(os.path.dirname(__file__), ".external_data"))

    # Download files
    downloaded_files = []
    for fileName, value in externals.items():
        algo, checksum = value
        print("fileName [%s] algo [%s] checksum [%s]" % (fileName, algo, checksum))
        downloaded_files.append(
            ExternalData(external_data_dir).download(
                uri=f"https://data.kitware.com/api/v1/file/hashsum/{algo}/{checksum}/download",
                fileName=fileName,
                checksum=f"{algo.upper()}:{checksum}",
            ),
        )

    # Copy download files to temporary directory
    for downloaded_file in downloaded_files:
        copyfile(downloaded_file, dest_dir.join(os.path.basename(downloaded_file)))

    return dest_dir



import pytest

from . import downloadExternals


@pytest.fixture
def external_data(request, tmpdir):
    """A pytest fixture to define a 'tmpdir' containing files or directories
    specified with a 'external_data' mark.
    """
    entry_list = []
    for mark in request.node.iter_markers('external_data'):
        entry_list.extend(mark.args)

    return downloadExternals(entry_list, tmpdir)



"""
The internal server API of Slicer Package Manager. Use these endpoints to
create new applications, new releases, and upload or download application and extensions
packages.
"""
import datetime
import re

from bson.objectid import ObjectId
from html_sanitizer import Sanitizer

from girder.api import access
from girder.constants import TokenScope, AccessType, SortDir
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import Resource
from girder.exceptions import RestException
from girder.models.folder import Folder
from girder.models.collection import Collection
from girder.utility import parseTimestamp

from ..models.extension import Extension as ExtensionModel
from ..models.package import Package as PackageModel
from .. import constants
from .. import utilities


class App(Resource):
    def __init__(self):
        super().__init__()
        self.resourceName = 'app'
        self._model = Folder()

        self.route('POST', (), self.initApp)
        self.route('GET', (), self.listApp)
        self.route('DELETE', (':app_id',), self.deleteApp)
        self.route('GET', (':app_id', 'downloadstats'), self.getDownloadStats)
        self.route('POST', (':app_id', 'release'), self.createNewRelease)
        self.route('GET', (':app_id', 'release'), self.getReleases)
        self.route('DELETE', (':app_id', 'release', ':release_id_or_name'),
                   self.deleteReleaseByIdOrName)
        self.route('POST', (':app_id', 'extension'), self.createOrUpdateExtension)
        self.route('GET', (':app_id', 'extension'), self.getExtensions)
        self.route('DELETE', (':app_id', 'extension', ':ext_id'), self.deleteExtension)
        self.route('POST', (':app_id', 'package'), self.createOrUpdatePackage)
        self.route('GET', (':app_id', 'package'), self.getPackages)
        self.route('DELETE', (':app_id', 'package', ':pkg_id'), self.deletePackage)
        self.route('GET', (':app_id', 'draft'), self.getAllDraftReleases)

    @autoDescribeRoute(
        Description('Create a new application.')
        .responseClass('Folder')
        .notes('If collectionId is missing or collectionName does not match an existing '
               'collection, a fresh new collection will be created with the "collection_name" '
               'given in parameters.<br/>'
               'By default the name "Applications" will be given to the collection.')
        .param('name', 'The name of the application.')
        .param('app_description', 'Application description.', required=False)
        .param('collection_id', 'The ID of the collection which contain the application',
               required=False)
        .param('collection_name', 'The Name of the collection which be created to contain'
               ' the application', required=False)
        .param('collection_description', 'Collection description.', required=False)
        .param('public', 'Whether the collection should be publicly visible.',
               required=False, dataType='boolean', default=True)
        .errorResponse('Write permission denied on the application.', 403),
    )
    @access.user(scope=TokenScope.DATA_WRITE)
    def initApp(self, name, app_description, collection_id, collection_name,
                collection_description, public):
        """
        Create the directory for start a new application. By default, without specifying
        a ``collection_id``, it will create a new collection name either ``collection_name``
        if provided, or **'Applications'**. If the collection 'Applications already exist it will
        get it.
        Return the new application (as a folder) that always contain a default
        sub-folder named 'draft'.

        :param name: Name of the new application
        :param app_description: Description of the new application
        :param collection_id: Id of the collection within create the application
        :param collection_name: Name of the collection which will be created
        :param collection_description: Description of the new collection
        :param public: Whether the new collection should be publicly visible
        :return: The new application folder
        """
        creator = self.getCurrentUser()
        # Load or create the collection that contain the application
        if collection_id:
            collection = Collection().load(collection_id, force=True)
        elif collection_name:
            collection = list(Collection().find({'name': collection_name}, user=creator))
            if not collection:
                collection = Collection().createCollection(
                    name=collection_name,
                    description=collection_description,
                    public=public,
                    creator=creator)
            else:
                collection = collection[0]
        else:
            collection = list(Collection().find({'name': 'Applications'}, user=creator))
            if not collection:
                collection = Collection().createCollection(
                    name='Applications',
                    description=collection_description,
                    public=public,
                    creator=creator)
            else:
                collection = collection[0]
        # Load or create the top_level_folder
        top_level_folder = list(self._model.childFolders(
            parent=collection,
            parentType='Collection',
            user=creator,
            filters={'name': constants.TOP_LEVEL_FOLDER_NAME}))
        if top_level_folder:
            top_level_folder = top_level_folder[0]
        else:
            # Create the 'package' folder
            top_level_folder = self._model.createFolder(
                parent=collection,
                name=constants.TOP_LEVEL_FOLDER_NAME,
                description='',
                parentType='Collection',
                public=public,
                creator=creator)

        # Create the application
        if not app_description:
            app_description = ''
        app = self._model.createFolder(
            parent=top_level_folder,
            name=name,
            description=app_description,
            parentType='Folder',
            public=public,
            creator=creator)
        # Create the 'draft' release which will be the default folder when uploading an extension
        self._model.createFolder(
            parent=app,
            name=constants.DRAFT_RELEASE_NAME,
            description='Uploaded each night, always up-to-date',
            parentType='Folder',
            public=public,
            creator=creator)
        # Set a default template name for extensions in the application,
        # this can be changed in anytime.
        return self._model.setMetadata(
            app,
            {
                'applicationPackageNameTemplate': constants.APPLICATION_PACKAGE_TEMPLATE_NAME,
                'extensionPackageNameTemplate': constants.EXTENSION_PACKAGE_TEMPLATE_NAME,
            },
        )

    @autoDescribeRoute(
        Description('List existing application.')
        .responseClass('Folder', array=True)
        .param('app_id', 'The ID of the application.', required=False)
        .param('collection_id', 'The ID of the collection.', required=False)
        .param('name', 'The name of the application.', required=False)
        .param('text', 'Provide text search of the application.', required=False)
        .pagingParams(defaultSort='name')
        .errorResponse()
        .errorResponse('Read permission denied on the application.', 403),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def listApp(self, app_id, collection_id, name, text, limit, offset, sort):
        """
        List existing applications base on some optional parameters.
        For searching application which aren't in the default collection (Applications),
        the parameter ``collection_id`` need to be provided.

        :param app_id: Application ID
        :param collection_id: Collection ID
        :param name: Name of the application
        :param text: Provide text search on the name of the application
        :return: List of applications
        """
        user = self.getCurrentUser()

        if ObjectId.is_valid(app_id):
            return self._model.load(app_id, user=user, level=AccessType.READ)
        else:
            if collection_id:
                parent = Collection().load(
                    collection_id, user=user, level=AccessType.READ, exc=True)
            else:
                parent = Collection().findOne(
                    query={'name': 'Applications'}, user=user, offset=offset)
            if parent:
                top_folder_name = list(self._model.childFolders(
                    parentType='collection', parent=parent, user=user,
                    filters={'name': constants.TOP_LEVEL_FOLDER_NAME}))
                if top_folder_name:
                    top_folder_name = top_folder_name[0]
                else:
                    return []
                filters = {}
                if text:
                    filters['$text'] = {
                        '$search': text,
                    }
                if name:
                    filters['name'] = name

                return list(self._model.childFolders(
                    parentType='Folder', parent=top_folder_name, user=user,
                    offset=offset, limit=limit, sort=sort, filters=filters))
            return []

    @access.user(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Delete an Application by ID.')
        .modelParam('app_id', destName='app_folder', model=Folder, level=AccessType.ADMIN)
        .param('progress', 'Whether to record progress on this task.',
               required=False, dataType='boolean', default=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin access was denied for the application.', 403),
    )
    def deleteApp(self, app_folder, progress):
        """
        Delete an application by ID.

        :param app_folder: Application folder loaded using ``app_id`` route parameter.
        :return: Confirmation message with the deleted application name
        """
        utilities.deleteFolder(app_folder, progress, self.getCurrentUser())
        return {'message': 'Deleted application %s.' % app_folder['name']}

    @autoDescribeRoute(
        Description('Create a new release.')
        .responseClass('Folder')
        .notes("The application's revision is stored as metadata of the new release.")
        .param('name', "The release's name.")
        .param('app_id', "The application's ID which contain the release", paramType='path')
        .param('app_revision', "The application's revision which correspond to the release")
        .param('description', "The application's description.", required=False)
        .param('public', 'Whether the release should be publicly visible.',
               required=False, dataType='boolean', default=True)
        .errorResponse(),
    )
    @access.user(scope=TokenScope.DATA_WRITE)
    def createNewRelease(self, name, app_id, app_revision, description, public):
        """
        Create a new release with the ``name`` within the application. The ``app_revision``
        will be used to automatically choose this release when uploading an application or
        extension package with a matching application revision metadata.

        :param name: Name of the new release
        :param app_id: Application ID
        :param app_revision: Revision of the application corresponding to this release
        :param description: Description of the new release
        :param public: Whether the release should be publicly visible

        :return: The new release folder
        """
        creator = self.getCurrentUser()
        application = self._model.load(app_id, user=creator, level=AccessType.READ)
        if not description:
            description = ''
        release = self._model.createFolder(
            parent=application,
            name=name,
            description=description,
            parentType='Folder',
            public=public,
            creator=creator)

        return self._model.setMetadata(release, {'revision': app_revision})

    @autoDescribeRoute(
        Description('Get all the releases from an application.')
        .responseClass('Folder')
        .param('app_id', "The application's ID.", paramType='path')
        .param('release_id_or_name', "The release's ID or name.", required=False)
        .pagingParams(defaultSort='created', defaultSortDir=SortDir.DESCENDING)
        .errorResponse('ID was invalid.')
        .errorResponse('Read permission denied on the application.', 403),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def getReleases(self, app_id, release_id_or_name, limit, offset, sort):
        """
        Get a list of all the stable release of an application. You can also search
        for a specific release if you provide the ``release_id_or_name`` parameter.
        If the ``release_id_or_name`` parameter doesn't correspond to any existing
        release, this will just return ``None``.

        :param app_id: Application ID
        :param release_id_or_name: Could be either the release ID or the release name
        :return: List of all release within the application or a specific release folder
        """
        user = self.getCurrentUser()
        application = self._model.load(app_id, user=user, level=AccessType.READ)

        if ObjectId.is_valid(release_id_or_name):
            return self._model.load(release_id_or_name, user=user, level=AccessType.READ)
        elif release_id_or_name:
            release_folder = list(self._model.childFolders(
                application,
                'Folder',
                user=user,
                filters={'lowerName': release_id_or_name.lower()}))
            if not release_folder:
                return None
            else:
                return release_folder[0]
        filters = {
            'name': {'$ne': constants.DRAFT_RELEASE_NAME},
        }
        return list(self._model.childFolders(
            application,
            'Folder',
            user=user,
            limit=limit,
            offset=offset,
            sort=sort,
            filters=filters))

    @autoDescribeRoute(
        Description('Get all the draft releases from an application.')
        .responseClass('Folder')
        .param('app_id', "The application's ID.", paramType='path')
        .param('revision', 'The revision of a draft release', required=False)
        .pagingParams(defaultSort='created', defaultSortDir=SortDir.DESCENDING)
        .errorResponse('ID was invalid.')
        .errorResponse('Read permission denied on the application.', 403),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def getAllDraftReleases(self, app_id, revision, limit, offset, sort):
        """
        Get a list of all the draft release of an application. It's also
        possible to filter this list by the metadata ``revision``.

        :param app_id: Application ID
        :param revision: Revision of one of the draft release
        :return: List of all release within the application
        """
        user = self.getCurrentUser()
        application = self._model.load(app_id, user=user, level=AccessType.READ)

        filters = {
            'name': constants.DRAFT_RELEASE_NAME,
        }
        release = list(self._model.childFolders(
            application,
            'Folder',
            user=user,
            filters=filters))
        if not release:
            msg = f"There is no {constants.DRAFT_RELEASE_NAME} release in this application."
            raise RestException(msg)
        release = release[0]
        draft_filters = {'meta.revision': {'$exists': True}}
        if revision:
            draft_filters = {'meta.revision': revision}
        return list(self._model.childFolders(
            release,
            'Folder',
            user=user,
            filters=draft_filters,
            limit=limit,
            offset=offset,
            sort=sort))

    @access.user(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Delete a release by ID or name.')
        .modelParam('app_id', destName='app_folder', model=Folder, level=AccessType.ADMIN)
        .param('release_id_or_name', "The release's ID or name.", paramType='path')
        .param('progress', 'Whether to record progress on this task.',
               required=False, dataType='boolean', default=False)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin access was denied for the release.', 403),
    )
    def deleteReleaseByIdOrName(self, app_folder, release_id_or_name, progress):
        """
        Delete a release by ID or name.

        :param app_folder: Application folder loaded using ``app_id`` route parameter.
        :param release_id_or_name: Could be either the release ID or the release name
        :param progress: Whether to record progress on this task
        :return: Confirmation message with the deleted release name
        """
        user = self.getCurrentUser()

        if ObjectId.is_valid(release_id_or_name):
            release = self._model.load(release_id_or_name, user=user)
        else:
            release_folder = list(self._model.childFolders(
                app_folder,
                'Folder',
                user=user,
                filters={'lowerName': release_id_or_name.lower()}))
            if not release_folder:
                release_folder = list(self._model.childFolders(
                    app_folder,
                    'Folder',
                    user=user,
                    filters={'lowerName': constants.DRAFT_RELEASE_NAME.lower()}))
                if not release_folder:
                    msg = f"Could not find the release '{release_id_or_name}'."
                    raise RestException(msg)
                revision_folder = list(self._model.childFolders(
                    release_folder[0],
                    'Folder',
                    user=user,
                    filters={'lowerName': release_id_or_name.lower()},
                ))
                if not revision_folder:
                    msg = f"Could not find the release '{release_id_or_name}'."
                    raise RestException(msg)
                release = revision_folder[0]
            else:
                release = release_folder[0]

        utilities.deleteFolder(release, progress, self.getCurrentUser())
        return {'message': 'Deleted release %s.' % release['name']}

    @autoDescribeRoute(
        Description('List or search available extensions.')
        .notes('If the "release_id" provided correspond to the "draft" release,'
               ' then you must provide the "app_revision" to use this parameters. '
               'If not, it will just be ignored.')
        .responseClass('Extension')
        .param('app_id', 'The ID of the application.', paramType='path')
        .param('extension_name', 'The name of the extension.', required=False)
        .param('release_id', 'The release id.', required=False)
        .param('extension_id', 'The extension id.', required=False)
        .param('os', 'The target operating system of the package.',
               required=False, enum=['linux', 'win', 'macosx'])
        .param('arch', 'The os chip architecture.',
               required=False, enum=['i386', 'amd64'])
        .param('app_revision', 'The revision of the application.', required=False)
        .param('baseName', 'The baseName of the extension', required=False)
        .param('q', 'The search query.', required=False)
        .pagingParams(defaultSort='created', defaultSortDir=SortDir.DESCENDING)
        .errorResponse(),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def getExtensions(self, app_id, extension_name, release_id, extension_id, os, arch,
                      app_revision, baseName, q, limit, sort, offset=0):
        """
        Get a list of extension which is filtered by some optional parameters. If the ``release_id``
        provided correspond to the draft release, then you must provide the app_revision to use
        this parameters. If not, it will just be ignored.

        :param app_id: Application ID
        :param extension_name: Extension name
        :param release_id: Release ID
        :param extension_id: Extension ID
        :param os: The operation system used for the extension.
        :param arch: The architecture compatible with the extension.
        :param app_revision: The revision of the application
        :param baseName: The baseName of the extension
        :param q: Text expected to be found in the extension name or description
        :return: The list of extensions
        """
        user = self.getCurrentUser()
        utilities.checkAccess(app_id, user)
        filters = {
            '$and': [
                {'meta.app_id': {'$eq': app_id}},
                {'meta.os': {'$exists': True}},
                {'meta.arch': {'$exists': True}},
                {'meta.app_revision': {'$exists': True}}],
        }
        if extension_name:
            filters['lowerName'] = extension_name.lower()
        if ObjectId.is_valid(extension_id):
            filters['_id'] = ObjectId(extension_id)
        if os:
            filters['meta.os'] = os
        if arch:
            filters['meta.arch'] = arch
        if app_revision:
            filters['meta.app_revision'] = app_revision
        if baseName:
            # Provide a exact match base on baseName
            filters['meta.baseName'] = baseName
        if q:
            escaped_query = re.escape(q)
            filters['$or'] = [
                {'meta.baseName': {'$regex': escaped_query, '$options': 'i'}},
                {'meta.description': {'$regex': escaped_query, '$options': 'i'}},
            ]
        if ObjectId.is_valid(release_id):
            release = self._model.load(release_id, user=user, level=AccessType.READ)
            if release['name'] == constants.DRAFT_RELEASE_NAME:
                if app_revision:
                    revisions = list(self._model.childFolders(
                        release,
                        'Folder',
                        user=user,
                        filters={'meta.revision': app_revision}))
                    if revisions:
                        extensions_folder = list(self._model.childFolders(
                            revisions[0],
                            'Folder',
                            user=user,
                            filters={'name': constants.EXTENSIONS_FOLDER_NAME}))
                        if extensions_folder:
                            filters['folderId'] = ObjectId(extensions_folder[0]['_id'])
                else:
                    revisions = self._model.childFolders(
                        release,
                        'Folder',
                        user=user,
                        sort=sort)
                    extensions = []
                    limit_tmp = limit
                    for revision in revisions:
                        extensions_folder = list(self._model.childFolders(
                            revision,
                            'Folder',
                            user=user,
                            filters={'name': constants.EXTENSIONS_FOLDER_NAME}))
                        # If the 'extensions' directory is not here, there are no extensions to list
                        if not extensions_folder:
                            continue
                        extensions_folder = extensions_folder[0]
                        filters['folderId'] = ObjectId(extensions_folder['_id'])
                        extensions += list(ExtensionModel().find(
                            query=filters,
                            limit=limit_tmp,
                            offset=offset,
                            sort=sort))
                        limit_tmp = limit - len(extensions)
                        if limit_tmp <= 0:
                            break
                    return extensions
            else:
                extensions_folder = list(self._model.childFolders(
                    release,
                    'Folder',
                    user=user,
                    filters={'name': constants.EXTENSIONS_FOLDER_NAME}))
                # If the 'extensions' directory is not here, there are no extensions to list
                if not extensions_folder:
                    return []
                extensions_folder = extensions_folder[0]
                filters['folderId'] = ObjectId(extensions_folder['_id'])
        return list(ExtensionModel().find(
            query=filters,
            limit=limit,
            offset=offset,
            sort=sort))

    @autoDescribeRoute(
        Description('Create or Update an extension package.')
        .param('app_id', 'The ID of the App.', paramType='path')
        .param('os', 'The target operating system of the package.',
               enum=['linux', 'win', 'macosx'])
        .param('arch', 'The os chip architecture.', enum=['i386', 'amd64'])
        .param('baseName', 'The baseName of the package (ie installer baseName).')
        .param('repository_type', 'The type of the repository (svn, git).')
        .param('repository_url', 'The url of the repository.')
        .param('revision', 'The svn or git revision of the extension.')
        .param('app_revision', 'The revision of the application '
                               'that the extension was built against.')
        .param('description', 'Text describing the extension.')
        .param('icon_url', 'The url of the icon for the extension.', required=False)
        .param('category', 'Category under which to place the extension. Subcategories should be '
               'delimited by character. If none is passed, will render under '
               'the Miscellaneous category..', required=False)

        .param('homepage', 'The url of the extension homepage.', required=False)
        .param('screenshots', 'Space-separate list of URLs of screenshots for the extension.',
               required=False)
        .param('contributors', 'List of contributors of the extension.', required=False)
        .param('dependency', 'List of the required extensions to use this one.', required=False)
        .param('license', 'The license short description of the extension.', required=False)
        .param('development_status', 'Arbitrary description of the status of the extension '
               '(stable, active, etc).', required=False)
        .param('enabled', 'Boolean indicating if the extension should be automatically enabled '
               'after its installation.', required=False)
        .errorResponse(),
    )
    @access.user(scope=TokenScope.DATA_WRITE)
    def createOrUpdateExtension(self, app_id, os, arch, baseName, repository_type, repository_url,
                                revision, app_revision, description,
                                icon_url, development_status, category, enabled, homepage,
                                screenshots, contributors, dependency, license):
        """
        Create or update an extension item.

        If a release identified by ``app_revision`` exists, it is used to create or lookup
        the corresponding release folder. Otherwise, a **draft** sub-folder named after
        ``app_revision`` is either used or created.

        If an extension with matching metadata ``meta.baseName``, ``meta.os``, ``meta.arch`` and
        ``meta.app_revision`` is found, the extension name and other metadata are updated.

        :param app_id: The ID of the application.
        :param os: The operation system used for the extension.
        :param arch: The architecture compatible with the extension.
        :param baseName: The base name of the extension.
        :param repository_type: The type of repository (github, gitlab, ...).
        :param repository_url: The Url of the repository.
        :param revision: The revision of the extension.
        :param app_revision: The revision of the application.
        :param description: The description of the extension.
        :return: The created/updated extension.
        """
        creator = self.getCurrentUser()
        application = Folder().load(app_id, user=creator)
        release_folder = utilities.getOrCreateReleaseFolder(
            application=application,
            user=creator,
            app_revision=app_revision)

        extensions_folder = list(self._model.childFolders(
            release_folder,
            'Folder',
            user=creator,
            filters={'name': constants.EXTENSIONS_FOLDER_NAME}))
        if not extensions_folder:
            extensions_folder = self._model.createFolder(
                parent=release_folder,
                name=constants.EXTENSIONS_FOLDER_NAME,
                parentType='Folder',
                public=release_folder['public'],
                description='This directory contains all the extensions packages',
                creator=creator)
        else:
            extensions_folder = extensions_folder[0]

        sanitizer = Sanitizer()
        description = sanitizer.sanitize(description)

        params = {
            'app_id': app_id,
            'baseName': baseName,
            'os': os,
            'arch': arch,
            'repository_type': repository_type,
            'repository_url': repository_url,
            'revision': revision,
            'app_revision': app_revision,
            'description': description,
        }
        if icon_url:
            params['icon_url'] = icon_url
        if development_status:
            params['development_status'] = development_status
        if category:
            params['category'] = category
        if enabled:
            params['enabled'] = enabled
        if homepage:
            params['homepage'] = homepage
        if screenshots:
            params['screenshots'] = screenshots
        if contributors:
            params['contributors'] = contributors
        if dependency:
            params['dependency'] = dependency
        if license:
            params['license'] = license

        name = application['meta']['extensionPackageNameTemplate'].format(**params)
        filters = {
            'meta.baseName': baseName,
            'meta.os': os,
            'meta.arch': arch,
            'meta.app_revision': app_revision,
        }
        # Only one extensions should be in this list
        extensions = list(ExtensionModel().get(extensions_folder, filters=filters))
        if not len(extensions):
            # The extension doesn't exist yet:
            extension = ExtensionModel().createExtension(name, creator, extensions_folder, params)
        elif len(extensions) == 1:
            # The extension already exist
            extension = extensions[0]
            # Check the file inside the extension Item
            files = ExtensionModel().childFiles(extension)
            if not files.count():
                # Extension empty
                msg = "Extension existing without any binary file."
                raise RestException(msg)

            # Update the extension
            extension['name'] = name
            extension = ExtensionModel().setMetadata(extension, params)
        else:
            msg = f"Too many extensions found for the same name '{name}'."
            raise RestException(msg)

        # Ready to upload the binary file
        return extension

    @autoDescribeRoute(
        Description('Delete an Extension by ID.')
        .param('app_id', 'The ID of the App.', paramType='path')
        .modelParam('ext_id', destName='ext_model', model=ExtensionModel, level=AccessType.WRITE)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin access was denied for the extension.', 403),
    )
    @access.user(scope=TokenScope.DATA_WRITE)
    def deleteExtension(self, app_id, ext_model):
        """
        Delete an extension by ID.

        :param app_id: Application ID
        :param ext_model: Extension model loaded using ``ext_id`` route parameter.
        :return: The deleted extension
        """
        user = self.getCurrentUser()
        utilities.checkAccess(app_id, user)
        ExtensionModel().remove(ext_model)
        return ext_model

    @autoDescribeRoute(
        Description('List or search available packages.')
        .notes('If the "release_id" provided correspond to the "draft" release,'
               ' then you must provide the "revision" to use this parameters. '
               'If not, it will just be ignored.')
        .responseClass('Package')
        .param('app_id', 'The ID of the application.', paramType='path')
        .param('package_name', 'The name of the package.', required=False)
        .param('release_id_or_name', "The release's ID or name", required=False)
        .param('package_id', 'The package id.', required=False)
        .param('os', 'The target operating system of the package.',
               required=False, enum=['linux', 'win', 'macosx'])
        .param('arch', 'The os chip architecture.',
               required=False, enum=['i386', 'amd64'])
        .param('revision', 'The revision of the application.', required=False)
        .param('baseName', 'The baseName of the package', required=False)
        .pagingParams(defaultSort='created', defaultSortDir=SortDir.DESCENDING)
        .errorResponse(),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def getPackages(self, app_id, package_name, release_id_or_name, package_id, os, arch,
                    revision, baseName, limit, offset, sort):
        """
        Get a list of package which is filtered by some optional parameters. If the ``release_id``
        provided correspond to the draft release, then you must provide the revision to use
        this parameters. If not, it will just be ignored.

        :param app_id: Application ID
        :param package_name: Package name
        :param release_id_or_name: Could be either the release ID or the release name
        :param package_id: Package ID
        :param os: The operation system used for the application package.
        :param arch: The architecture compatible with the application package.
        :param revision: The revision of the application
        :param baseName: The baseName of the package
        :return: The list of application packages
        """
        user = self.getCurrentUser()
        filters = {
            '$and': [
                {'meta.app_id': {'$eq': app_id}},
                {'meta.os': {'$exists': True}},
                {'meta.arch': {'$exists': True}},
                {'meta.revision': {'$exists': True}},
                {'meta.app_revision': {'$exists': False}}],
        }
        if package_name:
            filters['lowerName'] = package_name.lower()
        if ObjectId.is_valid(package_id):
            filters['_id'] = ObjectId(package_id)
        if os:
            filters['meta.os'] = os
        if arch:
            filters['meta.arch'] = arch
        if revision:
            filters['meta.revision'] = revision
        if baseName:
            # Provide a exact match base on baseName
            filters['meta.baseName'] = baseName

        release = None
        if ObjectId.is_valid(release_id_or_name):
            release = self._model.load(release_id_or_name, user=user, level=AccessType.READ)
        elif release_id_or_name:
            application = self._model.load(app_id, user=user, level=AccessType.READ)
            release_folder = list(self._model.childFolders(
                application,
                'Folder',
                user=user,
                filters={'lowerName': release_id_or_name.lower()}))
            if len(release_folder) > 0:
                release = release_folder[0]

        if release is not None:
            if release['name'] == constants.DRAFT_RELEASE_NAME:
                if revision:
                    revisions = list(self._model.childFolders(
                        release,
                        'Folder',
                        user=user,
                        filters={'meta.revision': revision}))
                    if revisions:
                        filters['folderId'] = ObjectId(revisions[0]['_id'])
                else:
                    revisions = self._model.childFolders(release, 'Folder', user=user)
                    packages = []
                    limit_tmp = limit
                    for rev in revisions:
                        filters['folderId'] = ObjectId(rev['_id'])
                        packages += list(PackageModel().find(
                            query=filters,
                            limit=limit_tmp,
                            offset=offset,
                            sort=sort))
                        if limit == 0:
                            continue
                        limit_tmp = limit - len(packages)
                        if limit_tmp <= 0:
                            break
                    return packages
            else:
                filters['folderId'] = ObjectId(release['_id'])

        return list(PackageModel().find(
            query=filters,
            limit=limit,
            offset=offset,
            sort=sort))

    @autoDescribeRoute(
        Description('Create or Update an application package.')
        .param('app_id', 'The ID of the App.', paramType='path')
        .param('os', 'The target operating system of the package.',
               enum=['linux', 'win', 'macosx'])
        .param('arch', 'The os chip architecture.', enum=['i386', 'amd64'])
        .param('baseName', 'The baseName of the package (ie installer baseName).')
        .param('repository_type', 'The type of the repository (svn, git).')
        .param('repository_url', 'The url of the repository.')
        .param('revision', 'The revision of the application')
        .param('version', 'The version of the application')
        .param('description', 'Text describing the package.', required=False)
        .param('pre_release', 'Boolean to specify if the package is a full release ready to be distributed.',
               dataType='boolean', required=False)
        .param('build_date', 'Build timestamp specified as a datetime string. Default set to current date and time.',
               required=False)
        .errorResponse(),
    )
    @access.user(scope=TokenScope.DATA_WRITE)
    def createOrUpdatePackage(self, app_id, os, arch, baseName, repository_type, repository_url,
                              revision, version, description, pre_release, build_date):
        """
        Create or update a package item.

        If a release identified by ``revision`` exists, it is used to create or lookup
        the corresponding release folder. Otherwise, a **draft** sub-folder named after
        ``revision`` is either used or created.

        If a package with matching metadata ``meta.baseName``, ``meta.os``, ``meta.arch`` and
        ``meta.revision`` is found, the package name and other metadata are updated.

        :param app_id: The ID of the application.
        :param os: The operation system used for the package.
        :param arch: The architecture compatible with the application package.
        :param baseName: The base name of the package.
        :param repository_type: The type of repository (github, gitlab, ...).
        :param repository_url: The Url of the repository.
        :param revision: The revision of the application.
        :param version: The version of the application.
        :param description: Description of the application package
        :param pre_release: Boolean to specify if the package is a full release ready to be distributed
        :param build_date: The build timestamp specified as datetime string. Default set to current date and time.
        :return: The created/updated package.
        """
        creator = self.getCurrentUser()
        application = Folder().load(app_id, user=creator)
        release_folder = utilities.getOrCreateReleaseFolder(
            application=application,
            user=creator,
            app_revision=revision)

        if not build_date:
            build_date = datetime.datetime.utcnow()
        else:
            try:
                build_date = parseTimestamp(build_date)
            except ValueError:
                msg = "Parameter 'build_date' is incorrectly formatted."
                raise RestException(msg)

        params = {
            'app_id': app_id,
            'baseName': baseName,
            'os': os,
            'arch': arch,
            'repository_type': repository_type,
            'repository_url': repository_url,
            'revision': revision,
            'version': version,
            'pre_release': pre_release,
            'build_date': build_date,
        }

        name = application['meta']['applicationPackageNameTemplate'].format(**params)
        filters = {
            'meta.baseName': baseName,
            'meta.os': os,
            'meta.arch': arch,
            'meta.revision': revision,
        }
        # Only one package should be in this list
        package = list(PackageModel().get(release_folder, filters=filters))
        if not len(package):
            # The package doesn't exist yet:
            package = PackageModel().createPackage(name, creator, release_folder, params,
                                                   description)
        elif len(package) == 1:
            # The package already exist
            package = package[0]
            # Check the file inside the application package
            files = PackageModel().childFiles(package)
            if not files.count():
                # Package empty
                msg = "Application Package existing without any binary file."
                raise RestException(msg)
            # Update the package
            package['name'] = name
            package = PackageModel().setMetadata(package, params)
        else:
            msg = f"Too many packages found for the same name '{name}'."
            raise RestException(msg)

        # Ready to upload the binary file
        return package

    @autoDescribeRoute(
        Description('Delete a Package by ID.')
        .param('app_id', 'The ID of the App.', paramType='path')
        .modelParam('pkg_id', destName='pkg_model', model=PackageModel, level=AccessType.WRITE)
        .errorResponse('ID was invalid.')
        .errorResponse('Admin access was denied for the package.', 403),
    )
    @access.user(scope=TokenScope.DATA_WRITE)
    def deletePackage(self, app_id, pkg_model):
        """
        Delete an application package by ID.

        :param app_id: Application ID
        :param pkg_model: Package model loaded using ``pkg_id`` route parameter.
        :return: The deleted package
        """
        user = self.getCurrentUser()
        utilities.checkAccess(app_id, user)
        PackageModel().remove(pkg_model)
        return pkg_model

    @autoDescribeRoute(
        Description('Get download stats of application and extensions packages '
                    'within an application.')
        .param('app_id', 'The ID of the application.', paramType='path')
        .errorResponse(),
    )
    @access.public(scope=TokenScope.DATA_READ)
    def getDownloadStats(self, app_id):
        """
        Get all the download count of all the application and extension packages
        from an application.

        :param app_id: Application ID
        :return: The JSON document of all the download statistics
        """
        user = self.getCurrentUser()
        application = self._model.load(app_id, user=user, level=AccessType.READ)
        releases = self._model.childFolders(application, 'Folder', user=user)

        downloadStats = {}

        for release in releases:
            if 'meta' in release and 'downloadStats' in release['meta']:
                if release['name'] == constants.DRAFT_RELEASE_NAME:
                    downloadStats.update(release['meta']['downloadStats'])
                else:
                    try:
                        downloadStats[release['meta']['revision']].update(
                            release['meta']['downloadStats'])
                    except KeyError:
                        downloadStats[
                            release['meta']['revision']] = release['meta']['downloadStats']

        return downloadStats



import json
import os

import pytest

from bson.objectid import ObjectId

from girder.models.collection import Collection
from girder.models.folder import Folder
from girder.models.file import File
from girder.models.item import Item
from girder.models.user import User

from pytest_girder.assertions import assertStatus, assertStatusOk
from pytest_girder.utils import getResponseBody

from slicer_package_manager import constants, utilities

from . import (
    computeFileChecksum,
    computeContentChecksum,
    downloadExternals,
    expectedDownloadStats,
    DRAFT_EXTENSIONS,
    DRAFT_PACKAGES,
    DRAFT_RELEASES,
    EXTENSIONS,
    FIXTURE_DIR,
    PACKAGES,
    RELEASE_EXTENSIONS,
    RELEASE_PACKAGES,
    RELEASES,
)


@pytest.fixture(name='user')
def fixture_user():
    return User().createUser('usr0', 'passwd', 'tst', 'usr', 'u@u.com')


@pytest.fixture(name='collection')
def fixture_collection(user):
    return Collection().createCollection(
        'testCollection',
        creator=user,
        description='Contain applications',
    )


@pytest.fixture(name='packages_folder')
def fixture_packages_folder(user, collection):
    return Folder().createFolder(
        parent=collection,
        name=constants.TOP_LEVEL_FOLDER_NAME,
        parentType='Collection',
        public=True,
        creator=user,
    )


@pytest.fixture(name='app_folder')
def fixture_app_folder(user, packages_folder):
    folder = Folder().createFolder(
        parent=packages_folder,
        name='application',
        description='app description',
        parentType='Folder',
        public=True,
        creator=user,
    )
    folder = Folder().setMetadata(
        folder,
        {
            'applicationPackageNameTemplate': constants.APPLICATION_PACKAGE_TEMPLATE_NAME,
            'extensionPackageNameTemplate': constants.EXTENSION_PACKAGE_TEMPLATE_NAME,
        },
    )
    return folder


@pytest.fixture(name='draft_release_folder')
def fixture_draft_release_folder(user, app_folder):
    return Folder().createFolder(
        parent=app_folder,
        name=constants.DRAFT_RELEASE_NAME,
        description='Uploaded each night, always up-to-date',
        parentType='Folder',
        public=True,
        creator=user,
    )


@pytest.fixture(name='draft_release_revision_folder')
def fixture_draft_release_revision_folder(user, draft_release_folder):
    folder = Folder().createFolder(
        parent=draft_release_folder,
        name=DRAFT_RELEASES[0]['revision'],
        parentType='Folder',
        public=True,
        creator=user,
    )
    folder = Folder().setMetadata(folder, {'revision': DRAFT_RELEASES[0]['revision']})
    return folder


@pytest.fixture(name='release_folder')
def fixture_release_folder(user, app_folder):
    folder = Folder().createFolder(
        parent=app_folder,
        name='release1',
        description='release description',
        parentType='Folder',
        public=True,
        creator=user,
    )
    folder = Folder().setMetadata(folder, {'revision': RELEASES[0]['revision']})
    return folder


@pytest.fixture(name='release_extensions')
def fixture_release_extensions(server, user, app_folder, release_folder, tmpdir, fsAssetstore):
    # Fix warnings related to fixtures not explicitly used.
    assert fsAssetstore

    downloadExternals(
        [os.path.join(FIXTURE_DIR, extension['filepath']) for extension in RELEASE_EXTENSIONS],
        tmpdir,
    )

    extensions = [_createOrUpdatePackage(
        server, 'extension', extension['meta'],
        filePath=tmpdir.join(extension['filepath']),
        _user=user, _app=app_folder,
    ) for extension in RELEASE_EXTENSIONS]

    for index, extension in enumerate(extensions):
        assert extension['name'] == RELEASE_EXTENSIONS[index]['name']
        isRelease = RELEASE_EXTENSIONS[index]['meta']['app_revision'] == release_folder['meta']['revision']
        assert isRelease
        extensions_folder = Folder().load(extension['folderId'], user=user)
        assert ObjectId(extensions_folder['parentId']) == release_folder['_id']

    return extensions


@pytest.fixture(name='draft_extensions')
def fixture_draft_extensions(server, user, app_folder, release_folder, draft_release_folder, tmpdir, fsAssetstore):
    # Fix warnings related to fixtures not explicitly used.
    assert draft_release_folder
    assert fsAssetstore

    downloadExternals(
        [os.path.join(FIXTURE_DIR, extension['filepath']) for extension in DRAFT_EXTENSIONS],
        tmpdir,
    )

    extensions = [_createOrUpdatePackage(
        server, 'extension', extension['meta'],
        filePath=tmpdir.join(extension['filepath']),
        _user=user, _app=app_folder,
    ) for extension in DRAFT_EXTENSIONS]

    for index, extension in enumerate(extensions):
        assert extension['name'] == DRAFT_EXTENSIONS[index]['name']
        isRelease = DRAFT_EXTENSIONS[index]['meta']['app_revision'] == release_folder['meta']['revision']
        assert not isRelease

    return extensions


@pytest.fixture(name='extensions')
def fixture_extensions(release_extensions, draft_extensions):
    extensions = []
    extensions.extend(release_extensions)
    extensions.extend(draft_extensions)
    return extensions


@pytest.fixture(name='release_packages')
def fixture_release_packages(server, user, app_folder, release_folder, tmpdir, fsAssetstore):
    # Fix warnings related to fixtures not explicitly used.
    assert fsAssetstore

    downloadExternals(
        [os.path.join(FIXTURE_DIR, package['filepath']) for package in RELEASE_PACKAGES],
        tmpdir,
    )
    packages = [_createOrUpdatePackage(
        server, 'package', package['meta'],
        filePath=tmpdir.join(package['filepath']),
        _user=user, _app=app_folder,
    ) for package in RELEASE_PACKAGES]

    for index, package in enumerate(packages):
        assert package['name'] == RELEASE_PACKAGES[index]['name']
        isRelease = RELEASE_PACKAGES[index]['meta']['revision'] == release_folder['meta']['revision']
        assert isRelease
        assert ObjectId(package['folderId']) == release_folder['_id']
        assert package['meta']['release'] == release_folder['name']

    return packages


@pytest.fixture(name='draft_packages')
def fixture_draft_packages(server, user, app_folder, release_folder, draft_release_folder, tmpdir, fsAssetstore):
    # Fix warnings related to fixtures not explicitly used.
    assert draft_release_folder
    assert fsAssetstore

    downloadExternals(
        [os.path.join(FIXTURE_DIR, package['filepath']) for package in DRAFT_PACKAGES],
        tmpdir,
    )
    packages = [_createOrUpdatePackage(
        server, 'package', package['meta'],
        filePath=tmpdir.join(package['filepath']),
        _user=user, _app=app_folder,
    ) for package in DRAFT_PACKAGES]

    for index, package in enumerate(packages):
        assert package['name'] == DRAFT_PACKAGES[index]['name']
        isRelease = DRAFT_PACKAGES[index]['meta']['revision'] == release_folder['meta']['revision']
        assert not isRelease
        assert 'release' not in package['meta']

    return packages


@pytest.fixture(name='packages')
def fixture_packages(release_packages, draft_packages):
    packages = []
    packages.extend(release_packages)
    packages.extend(draft_packages)
    return packages


@pytest.mark.plugin('slicer_package_manager')
def testInitApp(server, user, collection):
    _createApplicationCheck(
        server,
        'App_test',
        'Application without specifying any collection',
        collDescription='Automatic creation of the collection Applications',
        _user=user,
    )
    # Without any collection this should load the 'Applications' collection
    _createApplicationCheck(
        server,
        'App_test1',
        'Application without specifying any collection',
        _user=user,
    )
    # With a collection ID this should load the collection
    _createApplicationCheck(
        server,
        'App_test2',
        'Application with a collection ID',
        collId=collection['_id'],
        _user=user,
    )
    # With a collection name that match an existing collection name
    _createApplicationCheck(
        server,
        'App_test3',
        'Application with a collection name',
        collName=collection['name'],
        _user=user,
    )
    # With a collection name that does not exist yet
    _createApplicationCheck(
        server,
        'App_test4',
        'Application with a collection name',
        collName='Collection_for_app',
        _user=user,
    )


@pytest.mark.plugin('slicer_package_manager')
def testListApp(server, user, collection, app_folder):
    # Create applications in the 'Applications' collection
    app1 = _createApplicationCheck(
        server,
        'App_test1',
        'Application without specifying any collection',
        collDescription='Automatic creation of the collection Applications',
        _user=user,
    )
    app2 = _createApplicationCheck(
        server,
        'App_test2',
        'Application without specifying any collection',
        collDescription='Automatic creation of the collection Applications',
        _user=user,
    )

    # Get application with application ID
    resp = server.request(
        path='/app',
        method='GET',
        user=user,
        params={'app_id': app_folder['_id']},
    )
    assertStatusOk(resp)
    assert ObjectId(resp.json['_id']) == app_folder['_id']
    assert resp.json['name'] == app_folder['name']
    assert resp.json['description'] == app_folder['description']
    assert resp.json['meta']['applicationPackageNameTemplate'] == constants.APPLICATION_PACKAGE_TEMPLATE_NAME
    assert resp.json['meta']['extensionPackageNameTemplate'] == constants.EXTENSION_PACKAGE_TEMPLATE_NAME
    # List all applications from 'Applications' collection (Default)
    resp = server.request(
        path='/app',
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == 2
    assert resp.json[0]['_id'] == app1['_id']
    assert resp.json[1]['_id'] == app2['_id']
    # Get application with exact name
    resp = server.request(
        path='/app',
        method='GET',
        user=user,
        params={
            'collection_id': collection['_id'],
            'name': app_folder['name']},
    )
    assertStatusOk(resp)
    assert ObjectId(resp.json[0]['_id']) == app_folder['_id']
    assert resp.json[0]['name'] == app_folder['name']
    assert resp.json[0]['description'] == app_folder['description']
    assert resp.json[0]['meta']['applicationPackageNameTemplate'] == constants.APPLICATION_PACKAGE_TEMPLATE_NAME
    assert resp.json[0]['meta']['extensionPackageNameTemplate'] == constants.EXTENSION_PACKAGE_TEMPLATE_NAME


@pytest.mark.plugin('slicer_package_manager')
def testDeleteApp(server, user):
    # Create the application without any collection,
    # this should create the new 'Applications' collection
    app = _createApplicationCheck(
        server,
        'App_test',
        'Application without specifying any collection',
        collDescription='Automatic creation of the collection Applications',
        _user=user,
    )
    # Get the new application by ID
    resp = server.request(
        path='/app',
        method='GET',
        user=user,
        params={'app_id': app['_id']},
    )
    assertStatusOk(resp)
    assert resp.json['_id'] == app['_id']
    assert resp.json['name'] == app['name']
    # Delete the application
    resp = server.request(
        path='/app/%s' % app['_id'],
        method='DELETE',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json['message'] == f"Deleted application {app['name']}."
    # Try to get the application should failed
    resp = server.request(
        path='/app',
        method='GET',
        user=user,
        params={'app_id': app['_id']},
    )
    assertStatusOk(resp)
    assert resp.json is None


@pytest.mark.plugin('slicer_package_manager')
def testNewRelease(server, user, app_folder):
    _createReleaseCheck(
        server,
        name='V3.2.1',
        app_id=app_folder['_id'],
        app_revision='001',
        desc='This is a new release',
        _user=user,
    )


@pytest.mark.plugin('slicer_package_manager')
def testGetRelease(server, user, app_folder, release_folder, draft_release_folder):
    release1 = _createReleaseCheck(
        server,
        name='V3.2.1',
        app_id=app_folder['_id'],
        app_revision='002',
        desc='This is a new release',
        _user=user,
    )

    resp = server.request(
        path='/app/%s/release' % app_folder['_id'],
        method='GET',
        user=user,
    )
    # Check if it has return all the stable releases
    assertStatusOk(resp)
    assert len(resp.json) == 2
    assert resp.json[0]['_id'] == release1['_id']
    assert ObjectId(resp.json[1]['_id']) == release_folder['_id']

    # Get one release by ID
    resp = server.request(
        path='/app/%s/release' % app_folder['_id'],
        params={'release_id_or_name': release_folder['_id']},
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json['name'] == release_folder['name']
    assert resp.json['description'] == release_folder['description']
    assert ObjectId(resp.json['_id']) == release_folder['_id']

    # Get one release by name
    resp = server.request(
        path='/app/%s/release' % app_folder['_id'],
        params={'release_id_or_name': release_folder['name']},
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json['name'] == release_folder['name']
    assert resp.json['description'] == release_folder['description']
    assert ObjectId(resp.json['_id']) == release_folder['_id']

    resp = server.request(
        path='/app/%s/release' % app_folder['_id'],
        params={'release_id_or_name': constants.DRAFT_RELEASE_NAME},
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json['name'] == constants.DRAFT_RELEASE_NAME
    assert ObjectId(resp.json['_id']) == draft_release_folder['_id']


@pytest.mark.plugin('slicer_package_manager')
def testGetAllDraftRelease(server, user, app_folder, draft_release_revision_folder):
    resp = server.request(
        path='/app/%s/draft' % app_folder['_id'],
        method='GET',
        user=user,
    )
    # Check if it has return all the revision from the default release
    assertStatusOk(resp)
    assert len(resp.json) == 1
    assert ObjectId(resp.json[0]['_id']) == draft_release_revision_folder['_id']
    assert resp.json[0]['meta']['revision'] == draft_release_revision_folder['meta']['revision']
    # With the parameter 'revision'
    resp = server.request(
        path='/app/%s/draft' % app_folder['_id'],
        method='GET',
        user=user,
        params={'revision': draft_release_revision_folder['meta']['revision']},
    )
    # Check if it has return the good revision from the draft folder
    assertStatusOk(resp)
    assert len(resp.json) == 1
    assert ObjectId(resp.json[0]['_id']) == draft_release_revision_folder['_id']
    assert resp.json[0]['meta']['revision'] == draft_release_revision_folder['meta']['revision']
    # With the parameter 'revision' set to a wrong value
    resp = server.request(
        path='/app/%s/draft' % app_folder['_id'],
        method='GET',
        user=user,
        params={'revision': 'wrongRev'},
    )
    # Check if it has return the good revision from the draft folder
    assertStatusOk(resp)
    assert len(resp.json) == 0


@pytest.mark.plugin('slicer_package_manager')
def testDeleteReleaseByID(server, user, app_folder):
    _deleteRelease(server, '_id', _user=user, _app=app_folder)


@pytest.mark.plugin('slicer_package_manager')
def testDeleteReleaseByName(server, user, app_folder):
    _deleteRelease(server, 'name', _user=user, _app=app_folder)


@pytest.mark.plugin('slicer_package_manager')
def testDeleteRevisionRelease(server, user, app_folder, packages, extensions):
    # Fix warnings related to fixtures not explicitly used.
    assert packages
    assert extensions

    resp = server.request(
        path='/app/%s/draft' % app_folder['_id'],
        method='GET',
        user=user,
    )

    # Check if it has return all the revision from the default release
    assertStatusOk(resp)
    assert len(resp.json) == 2

    # Delete by Name the revision release '0001' in the "draft" release
    resp = server.request(
        path='/app/%s/release/%s' %
             (app_folder['_id'], EXTENSIONS[3]['meta']['app_revision']),
        method='DELETE',
        user=user,
    )
    assertStatusOk(resp)

    resp = server.request(
        path='/app/%s/draft' % app_folder['_id'],
        method='GET',
        user=user,
    )
    # Check if it has return all the revision from the default release
    assertStatusOk(resp)
    assert len(resp.json) == 1


@pytest.mark.plugin('slicer_package_manager')
def testUpdateExtensions(server, user, app_folder, extensions):
    # Update the same extension
    newParams = EXTENSIONS[1]['meta'].copy()
    newParams.update({
        'revision': '0000',
        'repository_type': 'gitlab',
        'description': (
            'Extension for Slicer 4 new version 2<br>'
            'with a new line and no script tag here:<script>alert(\'test\');</script>'
        ),
    })
    updatedExtension = _createOrUpdatePackage(server, 'extension', newParams, _user=user, _app=app_folder)
    # Check the same extension has different metadata
    assert updatedExtension['_id'] == extensions[1]['_id']
    assert updatedExtension['name'] == constants.EXTENSION_PACKAGE_TEMPLATE_NAME.format(**newParams)
    assert updatedExtension['meta'] != extensions[1]['meta']

    updatedExtensionMetadata = updatedExtension['meta']
    assert updatedExtensionMetadata['description'] == (
        'Extension for Slicer 4 new version 2<br>with a new line and no script tag here:'
    )


@pytest.mark.plugin('slicer_package_manager')
def testGetExtensions(server, user, app_folder, release_folder, extensions):
    # Get all the extension of the application
    resp = server.request(
        path='/app/%s/extension' % app_folder['_id'],
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == len(EXTENSIONS)

    # Get all the extension of the application for Linux
    resp = server.request(
        path='/app/%s/extension' % app_folder['_id'],
        method='GET',
        user=user,
        params={
            'os': 'linux',
        },
    )
    assertStatusOk(resp)
    assert len(resp.json) == 2
    for ext in resp.json:
        assert ext['meta']['os'] == 'linux'

    # Get all the extension of the application for Linux and amd64 architecture
    resp = server.request(
        path='/app/%s/extension' % app_folder['_id'],
        method='GET',
        user=user,
        params={
            'os': 'linux',
            'arch': 'amd64',
        },
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1
    for ext in resp.json:
        assert ext['meta']['os'] == 'linux'
        assert ext['meta']['arch'] == 'amd64'

    # Get all the extension of the application which are in "release1"
    resp = server.request(
        path='/app/%s/extension' % app_folder['_id'],
        method='GET',
        user=user,
        params={
            'release_id': release_folder['_id'],
        },
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1

    # Get all the extension of the application which are in the draft release
    draftRelease = list(Folder().childFolders(
        app_folder,
        'Folder',
        user=user,
        filters={'name': constants.DRAFT_RELEASE_NAME}))
    resp = server.request(
        path='/app/%s/extension' % app_folder['_id'],
        method='GET',
        user=user,
        params={
            'release_id': draftRelease[0]['_id'],
        },
    )
    assertStatusOk(resp)
    assert len(resp.json) == 4

    # Get a specific extension by name
    resp = server.request(
        path='/app/%s/extension' % app_folder['_id'],
        params={'extension_name': extensions[1]['name']},
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json[0]['_id'] == extensions[1]['_id']
    assert resp.json[0]['name'] == extensions[1]['name']

    # Get a specific extension with wrong name
    resp = server.request(
        path='/app/%s/extension' % app_folder['_id'],
        params={'extension_name': 'wrong_name'},
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == 0
    assert resp.json == []


@pytest.mark.plugin('slicer_package_manager')
def testDeleteExtensionPackages(server, user, app_folder, release_folder):
    # Create a new extension in the release "_release"
    extension = _createOrUpdatePackage(
        server,
        'extension',
        EXTENSIONS[0]['meta'],
        _user=user,
        _app=app_folder,
    )
    assert extension['name'] == EXTENSIONS[0]['name']
    extensions_folder = Folder().load(extension['folderId'], user=user)
    assert ObjectId(extensions_folder['parentId']) == release_folder['_id']
    # Get, delete, and try to re-get the package
    _deletePackages(server, 'extension', extension, _user=user, _app=app_folder)


@pytest.mark.external_data(
    os.path.join(FIXTURE_DIR, PACKAGES[2]['filepath']),
)
@pytest.mark.plugin('slicer_package_manager')
def testUploadAndDownloadPackages(server, user, app_folder, packages, external_data):
    # Try to create the same application package should just get the same one
    package2 = _createOrUpdatePackage(
        server,
        'package',
        PACKAGES[2]['meta'],
        external_data.join(PACKAGES[2]['filepath']),
        _user=user,
        _app=app_folder,
    )
    assert package2['name'] == packages[2]['name']


@pytest.mark.plugin('slicer_package_manager')
def testUpdatePackages(server, user, app_folder, packages):
    # Update the same package
    newParams = PACKAGES[1]['meta'].copy()
    newParams.update({
        'repository_url': 'https://AnotherURL.com',
        'repository_type': 'gitlab',
    })
    updatedPackage = _createOrUpdatePackage(server, 'package', newParams, _user=user, _app=app_folder)
    # Check the same package has different metadata
    assert updatedPackage['_id'] == packages[1]['_id']
    assert updatedPackage['name'] == constants.APPLICATION_PACKAGE_TEMPLATE_NAME.format(**newParams)
    assert updatedPackage['meta'] != packages[1]['meta']


@pytest.mark.plugin('slicer_package_manager')
def testGetPackages(server, user, app_folder, release_folder, packages):
    # Get all the package of the application
    resp = server.request(
        path='/app/%s/package' % app_folder['_id'],
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == 3

    # Get all the package of the application for Linux
    resp = server.request(
        path='/app/%s/package' % app_folder['_id'],
        method='GET',
        user=user,
        params={
            'os': 'linux',
        },
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1
    assert resp.json[0]['meta']['os'] == 'linux'

    # Get all the package of the application for Linux and amd64 architecture
    resp = server.request(
        path='/app/%s/package' % app_folder['_id'],
        method='GET',
        user=user,
        params={
            'os': 'macosx',
            'arch': 'i386',
        },
    )
    assertStatusOk(resp)
    assert len(resp.json) == 0

    # Get all the package of the application which are in "release1"
    resp = server.request(
        path='/app/%s/package' % app_folder['_id'],
        method='GET',
        user=user,
        params={
            'release_id_or_name': release_folder['_id'],
        },
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1

    # Get all the package of the application which are in the draft release
    draftRelease = list(Folder().childFolders(
        app_folder,
        'Folder',
        user=user,
        filters={'name': constants.DRAFT_RELEASE_NAME}))
    resp = server.request(
        path='/app/%s/package' % app_folder['_id'],
        method='GET',
        user=user,
        params={
            'release_id_or_name': draftRelease[0]['_id'],
        },
    )
    assertStatusOk(resp)
    assert len(resp.json) == 2

    # Get a specific package by name
    resp = server.request(
        path='/app/%s/package' % app_folder['_id'],
        params={'package_name': packages[1]['name']},
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json[0]['_id'] == packages[1]['_id']
    assert resp.json[0]['name'] == packages[1]['name']

    # Get a specific extension with wrong name
    resp = server.request(
        path='/app/%s/package' % app_folder['_id'],
        params={'package_name': 'wrong_name'},
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == 0
    assert resp.json == []


@pytest.mark.plugin('slicer_package_manager')
def testDeleteApplicationPackages(server, user, app_folder, release_folder):
    package = _createOrUpdatePackage(
        server,
        'package',
        PACKAGES[0]['meta'],
        _user=user,
        _app=app_folder,
    )
    assert package['name'] == PACKAGES[0]['name']
    assert ObjectId(package['folderId']) == release_folder['_id']

    # Get, delete, and try to re-get the package
    _deletePackages(server, 'package', package, _user=user, _app=app_folder)


@pytest.mark.plugin('slicer_package_manager')
def testDownloadStats(server, user, app_folder, draft_release_revision_folder, packages, extensions):
    # Fix warnings related to fixtures not explicitly used.
    assert packages

    # Get the downloadStats
    expectedStats = expectedDownloadStats
    resp = server.request(
        path='/app/%s/downloadstats' % app_folder['_id'],
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json == expectedStats

    # Download multiple time an extension
    ext3_file = list(File().find({'itemId': ObjectId(extensions[3]['_id'])}))
    ext4_file = list(File().find({'itemId': ObjectId(extensions[4]['_id'])}))

    N = 5
    for _idx in range(N):
        _downloadFile(server, ext3_file[0]['_id'], _user=user)
        _downloadFile(server, ext4_file[0]['_id'], _user=user)

    expectedStats['0001']['extensions']['Ext2']['macosx'].update({
        'amd64': N + expectedStats['0001']['extensions']['Ext2']['macosx']['amd64'],
        'i386': N + expectedStats['0001']['extensions']['Ext2']['macosx']['i386'],
    })
    resp = server.request(
        path='/app/%s/downloadstats' % app_folder['_id'],
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json == expectedStats

    # The download stats aren't affected by deletion of extension or revision
    # in the default release
    # Delete extensions
    resp = server.request(
        path='/app/%(app_id)s/extension/%(ext_id)s' % {
            'app_id': app_folder['_id'],
            'ext_id': extensions[3]['_id']},
        method='DELETE',
        user=user,
    )
    assertStatusOk(resp)
    resp = server.request(
        path='/app/%(app_id)s/extension/%(ext_id)s' % {
            'app_id': app_folder['_id'],
            'ext_id': extensions[4]['_id']},
        method='DELETE',
        user=user,
    )
    assertStatusOk(resp)

    # Delete the revision '0000' in the "draft" release
    resp = server.request(
        path='/app/%(app_id)s/release/%(release_id)s' % {
            'app_id': app_folder['_id'],
            'release_id': draft_release_revision_folder['_id']},
        method='DELETE',
        user=user,
    )
    assertStatusOk(resp)

    # Check if download Stats are the same
    resp = server.request(
        path='/app/%s/downloadstats' % app_folder['_id'],
        method='GET',
        user=user,
    )
    assertStatusOk(resp)
    assert resp.json == expectedStats


@pytest.mark.plugin('slicer_package_manager')
def testGetReleaseFolder(server, user, release_folder, packages, extensions):
    # Fix warnings related to fixtures not explicitly used.
    assert server

    # Release package
    release_package = packages[0]
    assert utilities.getReleaseFolder(release_package)['_id'] == release_folder['_id']

    # Draft package
    draft_package = packages[1]
    assert utilities.getReleaseFolder(draft_package)['name'] == constants.DRAFT_RELEASE_NAME

    # Release extension
    release_extension = extensions[0]
    extensions_folder = Folder().load(release_extension['folderId'], user=user)
    assert ObjectId(extensions_folder['parentId']) == release_folder['_id']
    assert utilities.getReleaseFolder(release_extension)['_id'] == release_folder['_id']

    # Draft extension
    draft_extension = extensions[1]
    assert utilities.getReleaseFolder(draft_extension)['name'] == constants.DRAFT_RELEASE_NAME


@pytest.mark.plugin('slicer_package_manager')
def testIsApplicationFolder(server, app_folder, release_folder, draft_release_folder, draft_release_revision_folder):
    # Fix warnings related to fixtures not explicitly used.
    assert server

    assert utilities.isApplicationFolder(app_folder)
    assert not utilities.isApplicationFolder(release_folder)
    assert not utilities.isApplicationFolder(draft_release_folder)
    assert not utilities.isApplicationFolder(draft_release_revision_folder)


@pytest.mark.plugin('slicer_package_manager')
def testIsReleaseFolder(server, app_folder, release_folder, draft_release_folder, draft_release_revision_folder):
    # Fix warnings related to fixtures not explicitly used.
    assert server

    assert not utilities.isReleaseFolder(app_folder)
    assert utilities.isReleaseFolder(release_folder)
    assert not utilities.isReleaseFolder(draft_release_folder)
    assert utilities.isReleaseFolder(draft_release_revision_folder)


@pytest.mark.plugin('slicer_package_manager')
def testIsDraftReleaseFolder(server, app_folder, release_folder, draft_release_folder, draft_release_revision_folder):
    # Fix warnings related to fixtures not explicitly used.
    assert server

    assert not utilities.isDraftReleaseFolder(app_folder)
    assert not utilities.isDraftReleaseFolder(release_folder)
    assert not utilities.isDraftReleaseFolder(draft_release_folder)
    assert utilities.isDraftReleaseFolder(draft_release_revision_folder)


@pytest.mark.parametrize(
    ('action', 'method', 'src_folder', 'dest_folder', 'items', 'release_before', 'release_after'),
    [
        (
            'copy',
            'POST',
            pytest.lazy_fixture('draft_release_revision_folder'),
            pytest.lazy_fixture('release_folder'),
            pytest.lazy_fixture('draft_packages'),
            None,
            RELEASES[0]['name'],
        ),
        (
            'copy',
            'POST',
            pytest.lazy_fixture('release_folder'),
            pytest.lazy_fixture('draft_release_revision_folder'),
            pytest.lazy_fixture('release_packages'),
            RELEASES[0]['name'],
            None,
        ),
        (
            'move',
            'PUT',
            pytest.lazy_fixture('draft_release_revision_folder'),
            pytest.lazy_fixture('release_folder'),
            pytest.lazy_fixture('draft_packages'),
            None,
            RELEASES[0]['name'],
        ),
        (
            'move',
            'PUT',
            pytest.lazy_fixture('release_folder'),
            pytest.lazy_fixture('draft_release_revision_folder'),
            pytest.lazy_fixture('release_packages'),
            RELEASES[0]['name'],
            None,
        ),
    ],
    ids=[
        'copy_package_from_draft_to_release',
        'copy_package_from_release_to_draft',
        'move_package_from_draft_to_release',
        'move_package_from_release_to_draft',
    ],
)
@pytest.mark.plugin('slicer_package_manager')
def testApplicationPackageMetadataAutoUpdate(
        server, user, action, method, src_folder, dest_folder, items, release_before, release_after):

    item = items[0]

    assert ObjectId(item['folderId']) == src_folder['_id']
    assert item['meta'].get('release', None) == release_before

    resp = server.request(
        path='/resource/%s' % action,
        method=method,
        user=user,
        params={
            'resources': json.dumps({'item': [item['_id']]}),
            'parentType': 'folder',
            'parentId': dest_folder['_id'],
        },
    )
    assertStatusOk(resp)

    resp = server.request(
        path='/item',
        method='GET',
        user=user,
        params={
            'folderId': dest_folder['_id'],
            'name': item['name'],
        },
    )
    assertStatusOk(resp)
    item_after = resp.json[0]

    for key in ['name', 'size']:
        assert item_after[key] == item[key]

    for key in ['app_id', 'arch', 'baseName', 'os', 'revision']:
        assert item_after['meta'][key] == item['meta'][key]

    assert ObjectId(item_after['folderId']) == dest_folder['_id']
    assert item_after['meta'].get('release', None) == release_after


@pytest.mark.parametrize("items", [
    pytest.lazy_fixture('packages'),
    pytest.lazy_fixture('extensions'),
])
@pytest.mark.plugin('slicer_package_manager')
def testPackageMetadataChecksumUpdate_1(server, user, items):
    """Replace existing file and verify checksum is updated."""
    item = Item().load(items[0]['_id'], force=True)
    file = list(Item().childFiles(item))[0]

    # Replace file
    updated_contents = "you've changed!"
    _replaceFile(server, file["_id"], updated_contents, user)

    # Confirm item checksum was updated
    item_after = Item().load(item['_id'], force=True)
    expected_checksum = computeContentChecksum("SHA512", updated_contents.encode("utf8"))
    assert item_after["meta"]["sha512"] == expected_checksum


@pytest.mark.parametrize("items", [
    pytest.lazy_fixture('packages'),
    pytest.lazy_fixture('extensions'),
])
@pytest.mark.plugin('slicer_package_manager')
def testPackageMetadataChecksumUpdate_2(server, user, items, tmpdir):
    """Upload an additional file and verify checksum remains the same."""
    item = Item().load(items[1]['_id'], force=True)
    assert Item().childFiles(item).count() == 1
    item_first_file_sha512 = item["meta"]["sha512"]

    # Upload additional file
    filePath = tmpdir / "additional.tar.gz"
    with open(filePath, "w") as content:
        content.write("new file contents")
    additional_file_sha512 = computeFileChecksum("SHA512", filePath)
    assert item_first_file_sha512 != additional_file_sha512
    with open(filePath) as content:
        server.uploadFile(filePath, content.read(), user, item, parentType="item")

    # Check the additional file was uploaded
    item_after = Item().load(item['_id'], force=True)
    assert Item().childFiles(item_after).count() == 2

    # Confirm item checksum is unchanged
    assert item_first_file_sha512 == item_after["meta"]["sha512"]


@pytest.mark.parametrize("items", [
    pytest.lazy_fixture('packages'),
    pytest.lazy_fixture('extensions'),
])
@pytest.mark.plugin('slicer_package_manager')
def testPackageMetadataChecksumUpdate_3(server, user, items, tmpdir):
    """Upload an additional file, remove the first one and verify the checksum matches
    the one of the additional file.
    """
    item = Item().load(items[2]['_id'], force=True)
    assert Item().childFiles(item).count() == 1
    item_first_file_sha512 = item["meta"]["sha512"]
    item_first_file_id = list(Item().childFiles(item))[0]["_id"]

    # Upload additional file
    filePath = tmpdir / "additional.tar.gz"
    with open(filePath, "w") as content:
        content.write("new file contents")
    additional_file_sha512 = computeFileChecksum("SHA512", filePath)
    assert item_first_file_sha512 != additional_file_sha512
    with open(filePath) as content:
        server.uploadFile(filePath, content.read(), user, item, parentType="item")

    # Check the additional file was uploaded
    item_after = Item().load(item['_id'], force=True)
    assert Item().childFiles(item_after).count() == 2

    # Remove the first file
    resp = server.request(
        path='/file/%s' % item_first_file_id,
        method='DELETE',
        user=user,
    )
    assertStatusOk(resp)
    item_after = Item().load(item['_id'], force=True)
    assert Item().childFiles(item_after).count() == 1

    # Confirm the item checksum was updated
    assert item_after["meta"]["sha512"] != item_first_file_sha512
    assert item_after["meta"]["sha512"] == additional_file_sha512


@pytest.mark.parametrize("items", [
    pytest.lazy_fixture('packages'),
    pytest.lazy_fixture('extensions'),
])
@pytest.mark.plugin('slicer_package_manager')
def testPackageMetadataChecksumUpdate_4(server, user, items):
    """Delete the last file and verify the checksum is set to an empty string."""
    item = Item().load(items[0]['_id'], force=True)
    assert Item().childFiles(item).count() == 1
    item_file_id = list(Item().childFiles(item))[0]["_id"]

    # Remove the file
    resp = server.request(
        path='/file/%s' % item_file_id,
        method='DELETE',
        user=user,
    )
    assertStatusOk(resp)
    item_after = Item().load(item['_id'], force=True)
    assert Item().childFiles(item_after).count() == 0

    # Confirm the item checksum was updated
    assert not item_after["meta"]["sha512"]


@pytest.mark.parametrize(
    ('build_date', 'expected_build_date', 'status_code'), [
        (None, None, 200),
        ('2021-06-21 22:00:26', '2021-06-21T22:00:26+00:00', 200),
        ('2021-06-21T22:00:26+0000', '2021-06-21T22:00:26+00:00', 200),
        ('2021-06-21T22:00:26+00:00', '2021-06-21T22:00:26+00:00', 200),
        ('2021-06-21 11:37:36 -0400', '2021-06-21T15:37:36+00:00', 200),
        ('2021-06-23 02:54:11-04:00', '2021-06-23T06:54:11+00:00', 200),
        ('abcdef', None, 400),
    ],
    ids=[
        'default',
        'no-timezone',
        'timezone-without-colon',
        'timezone-with-colon',
        'git-date-iso-local',
        'date-rfc-3339-seconds',
        'invalid',
    ],
)
@pytest.mark.plugin('slicer_package_manager')
def testMetadataBuildDate(build_date, expected_build_date, status_code, server, user, app_folder, draft_release_folder):
    # Fix warnings related to fixtures not explicitly used.
    assert draft_release_folder

    newParams = PACKAGES[0]['meta'].copy()
    if build_date is not None:
        newParams['build_date'] = build_date
    package = _createOrUpdatePackage(
        server,
        'package',
        newParams,
        _user=user,
        _app=app_folder,
        status_code=status_code,
    )
    if status_code != 200:
        return
    assert package['name'] == PACKAGES[0]['name']
    assert package['meta'].get('build_date')
    if expected_build_date is not None:
        assert package['meta'].get('build_date') == expected_build_date


def _createApplicationCheck(server, appName, appDescription, collId=None,
                            collName=None, collDescription='', _user=None):
    params = {
        'name': appName,
        'app_description': appDescription,
        'collection_description': collDescription,
    }
    if not (collId or collName):
        collName = 'Applications'
        params.update({'collection_name': collName})
    elif collId:
        if not ObjectId.is_valid(collId):
            raise Exception("The ObjectId %s is not valid" % collId)
        params.update({'collection_id': collId})
        collection = Collection().load(collId, user=_user)
        collName = collection['name']
    else:
        params.update({'collection_name': collName})

    resp = server.request(
        path='/app',
        method='POST',
        user=_user,
        params=params,
    )
    # Check if it has created the application
    assertStatusOk(resp)
    assert resp.json['name'] == appName
    assert resp.json['description'] == appDescription
    assert resp.json['meta']['applicationPackageNameTemplate'] == constants.APPLICATION_PACKAGE_TEMPLATE_NAME
    assert resp.json['meta']['extensionPackageNameTemplate'] == constants.EXTENSION_PACKAGE_TEMPLATE_NAME
    # Check if it has created/load the collection
    topLevelFolder = Folder().load(resp.json['parentId'], user=_user)
    collection = Collection().load(topLevelFolder['parentId'], user=_user)
    assert collection['name'] == collName
    if collDescription:
        assert collection['description'] == collDescription
    # Check if it has created the "draft" release
    parent = Folder().load(resp.json['_id'], user=_user)
    draft = list(Folder().childFolders(parent, 'Folder', user=_user))
    assert len(draft) == 1
    assert draft[0]['name'] == constants.DRAFT_RELEASE_NAME
    assert draft[0]['description'] == 'Uploaded each night, always up-to-date'

    return resp.json


def _createReleaseCheck(server, name, app_id, app_revision, desc='', _user=None):
    resp = server.request(
        path='/app/%s/release' % app_id,
        method='POST',
        user=_user,
        params={
            'name': name,
            'app_revision': app_revision,
            'description': desc,
        },
    )
    # Check if it has created the new release
    assertStatusOk(resp)
    assert resp.json['name'] == name
    assert resp.json['description'] == desc
    assert resp.json['meta']['revision'] == app_revision
    return resp.json


def _deleteRelease(server, identifier, _user=None, _app=None):
    # Create the release
    release = _createReleaseCheck(
        server,
        name='V3.2.1',
        app_id=_app['_id'],
        app_revision='001',
        desc='This is a new release',
        _user=_user,
    )
    # Get the new release by name
    resp = server.request(
        path='/app/%s/release' % _app['_id'],
        params={'release_id_or_name': release[identifier]},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json['_id'] == release['_id']
    assert resp.json['name'] == release['name']
    # Delete the release by name
    resp = server.request(
        path='/app/%(app_id)s/release/%(id_or_name)s' % {
            'app_id': _app['_id'],
            'id_or_name': release[identifier]},
        method='DELETE',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json['message'] == f"Deleted release {release['name']}."
    # Try to get the release should failed
    resp = server.request(
        path='/app/%s/release' % _app['_id'],
        params={'release_id_or_name': release[identifier]},
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json is None


# def _uploadExternalData(server, item, filePath, fileName=None, _user=None):
#     file = os.path.join(self.dataDir, filePath)
#     size = os.path.getsize(file)
#     offset = 0
#     if fileName is None:
#         fileName = filePath
#
#     # Initialize the upload
#     resp = server.request(
#         path='/file', method='POST', user=_user, params={
#             'parentType': 'item',
#             'parentId': item['_id'],
#             'name': fileName,
#             'size': size,
#             'mimeType': 'application/octet-stream'
#         })
#     assertStatusOk(resp)
#     uploadId = resp.json['_id']
#
#     # Upload content
#     with open(file, 'rb') as fp:
#         while True:
#             chunk = fp.read(min(self._MAX_CHUNK_SIZE, (size - offset)))
#             if not chunk:
#                 break
#             if isinstance(chunk, str):
#                 chunk = chunk.encode('utf8')
#             resp = server.request(
#                 path='/file/chunk',
#                 method='POST',
#                 user=_user,
#                 params={
#                     'offset': offset,
#                     'uploadId': uploadId
#                 },
#                 body=chunk,
#                 type='application/octet-stream'
#             )
#             assertStatusOk(resp)
#             offset += len(chunk)
#
#         uploadedFile = resp.json
#         assert 'itemId' in uploadedFile
#         assert uploadedFile['assetstoreId'] == str(self.assetstore['_id'])
#         assert uploadedFile['name'] == filePath
#         assert uploadedFile['size'] == size
#     return uploadedFile


def _downloadFile(server, _id, _user=None):
    resp = server.request(
        path="/file/%s/download" % _id,
        method='GET',
        user=_user,
        isJson=False)
    assertStatusOk(resp)
    return getResponseBody(resp, text=False)


def _replaceFile(server, _id, contents, _user):

    upload = server.request(
        path='/file/%s/contents' % _id,
        method='PUT',
        user=_user,
        params={
            'size': len(contents),
        })
    assertStatusOk(upload)

    resp = server.request(
        path='/file/chunk',
        method='POST',
        user=_user,
        body=contents,
        params={
            'uploadId': upload.json['_id'],
            'offset': 0,
        },
        type='text/plain')
    assertStatusOk(resp)
    return getResponseBody(resp, text=False)


def _createOrUpdatePackage(server, packageType, params, filePath=None, _user=None, _app=None, status_code=200):
    resp = server.request(
        path='/app/%s/%s' % (_app['_id'], packageType),
        method='POST',
        user=_user,
        params=params)
    assertStatus(resp, status_code)

    if status_code != 200:
        return None

    def _filtered(metadata):
        return {k: v for (k, v) in metadata.items() if k not in (
            'app_id',
            'build_date',
            'description',
            'sha512',
        )}

    # assert every other fields (besides unique, computed or sanitized ones) are identical
    assert _filtered(resp.json['meta']) == _filtered(params)

    if filePath:
        # Upload a package file
        item = resp.json
        with open(filePath, 'rb') as content:
            uploadedFile = server.uploadFile(
                filePath, content.read(), _user, item, parentType="item")

        # Download the package and compare checksum
        downloaded_checksum = computeContentChecksum('SHA512', _downloadFile(server, uploadedFile['_id'], _user=_user))
        local_checksum = computeFileChecksum('SHA512', filePath)
        assert downloaded_checksum == local_checksum

    resp = server.request(
        path='/app/%s/%s' % (_app['_id'], packageType),
        method='GET',
        user=_user,
        params={'%s_id' % packageType: resp.json['_id']},
    )
    assertStatusOk(resp)

    assert len(resp.json) == 1

    if filePath:
        item = resp.json[0]
        assert item['meta']['sha512'] == local_checksum

    return resp.json[0]


def _deletePackages(server, packageType, pkg, _user=None, _app=None):
    # Get all the  application package, this should only be the new one
    resp = server.request(
        path='/app/%s/%s' % (_app['_id'], packageType),
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == 1
    assert resp.json[0]['_id'] == pkg['_id']
    # Delete the package
    resp = server.request(
        path='/app/%(app_id)s/%(type)s/%(pkg_id)s' % {
            'app_id': _app['_id'],
            'type': packageType,
            'pkg_id': pkg['_id']},
        method='DELETE',
        user=_user,
    )
    assertStatusOk(resp)
    assert resp.json['_id'] == pkg['_id']
    # Try to get the package shouldn't success
    resp = server.request(
        path='/app/%s/%s' % (_app['_id'], packageType),
        method='GET',
        user=_user,
    )
    assertStatusOk(resp)
    assert len(resp.json) == 0



"""
Constants should be defined here.
"""

TOP_LEVEL_FOLDER_NAME = 'packages'
DRAFT_RELEASE_NAME = 'draft'
APPLICATION_PACKAGE_TEMPLATE_NAME = '{baseName}_{os}_{arch}_{revision}'
EXTENSION_PACKAGE_TEMPLATE_NAME = '{app_revision}_{baseName}_{os}_{arch}_{revision}'
EXTENSIONS_FOLDER_NAME = 'extensions'



import datetime

from girder.models.folder import Folder
from girder.models.item import Item
from girder.models.model_base import ValidationException


class Extension(Item):
    """
    The ``Extension`` class derives from the ``Item`` model in Girder and it embeds
    extension binary files.
    """

    def initialize(self):
        super().initialize()
        # To be able to upload within an Extension the name has to stay as 'item'.
        self.name = 'item'

    def get(self, release, limit=0, offset=0, sort=None, filters=None, **kwargs):
        """
        Get all the extensions available for a release.

        :param release: The release folder
        :return: Generator containing all the child items of the release.
        """
        return Folder().childItems(
            release,
            limit=limit,
            offset=offset,
            sort=sort,
            filters=filters,
            **kwargs)

    def createExtension(self, name, creator, folder, params):
        """
        Create and save in the DB a new extension.

        :param name: Name of the new extension
        :param creator: The creator user
        :param folder: The release folder within the extension will be created.
        :param params: All the metadata to set on the new extension
        :return: The new extension item
        """
        item = self.createItem(
            name,
            creator,
            folder,
            params.get('description'),
        )
        return self.setMetadata(item, params)

    def validate(self, doc):
        """
        Validate the extension instance.

        :param doc: The extension instance
        :return: The extension instance once validated
        """
        # Call Item validate method
        doc = super().validate(doc)

        if not isinstance(doc.get('created'), datetime.datetime):
            msg = 'Extension field "created" must be a datetime.'
            raise ValidationException(msg)

        # Validate the meta field
        if doc.get('meta'):
            base_params = {
                'app_id',
                'os',
                'arch',
                'revision',
                'repository_type',
                'repository_url',
                'app_revision',
                'baseName',
                'description',
                'sha512',
            }
            specs = []
            for meta in base_params:
                specs.append({
                    'name': meta,
                    'type': (str,),
                    'exception_msg': f'Extension field "{meta}" must be a non-empty string.',
                })
            for spec in specs:
                if doc['meta'].get(spec['name']) and not isinstance(
                   doc['meta'][spec['name']], spec['type']):
                    msg=spec['exception_msg']
                    raise ValidationException(msg)
            extraMeta = set(doc['meta'].keys()) - base_params
            if extraMeta:
                extra_params = {
                    'icon_url',
                    'development_status',
                    'category',
                    'enabled',
                    'homepage',
                    'screenshots',
                    'contributors',
                    'dependency',
                    'license',
                }
                if extraMeta - extra_params:
                    msg = f'Extension has extra fields: {", ".join(sorted(extraMeta))}.'
                    raise ValidationException(msg)
                specs = []
                for meta in extra_params:
                    if meta == 'enabled':
                        specs.append({
                            'name': meta,
                            'type': bool,
                            'exception_msg': f'Extension field "{meta}" must be a boolean.',
                        })
                    else:
                        specs.append({
                            'name': meta,
                            'type': (str,),
                            'exception_msg': f'Extension field "{meta}" must be a non-empty string.',
                        })
                for spec in specs:
                    if doc['meta'].get(spec['name']) and not isinstance(
                       doc['meta'][spec['name']], spec['type']):
                        msg = spec['exception_msg']
                        raise ValidationException(msg)
            duplicateQuery = {
                'name': doc['name'],
                'os': doc['meta']['os'],
                'arch': doc['meta']['arch'],
                'app_revision': doc['meta']['app_revision'],
            }
            if '_id' in doc:
                duplicateQuery['_id'] = {'$ne': doc['_id']}
            if self.findOne(duplicateQuery, fields=['_id']):
                msg = 'An Extension with this name and characteristics already exists.'
                raise ValidationException(msg)
        return doc



import versioneer

from setuptools import setup

# This file is required to support editable install
# See https://setuptools.pypa.io/en/latest/userguide/pyproject_config.html

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



import pytest

from girder.plugin import loadedPlugins


@pytest.mark.plugin('slicer_package_manager')
def test_import(server):
    assert server  # Fix warnings related to fixtures not explicitly used.

    assert 'slicer_package_manager' in loadedPlugins()



from girder.constants import AccessType
from girder.models.folder import Folder
from girder.models.item import Item
from girder.utility.progress import ProgressContext

from . import constants


def isSlicerPackages(item):
    """Return True if the item represents either an application or an extension package."""
    if 'meta' in item and all(k in item['meta'] for k in ('os', 'arch', 'baseName', 'revision')):
        return True
    return False


def isChildOfSlicerPackages(file_item):
    """
    Return True if the file is included in an item corresponding to either an application or
    an extension package.
    """
    parent_item = Item().load(file_item['itemId'], force=True)
    return isSlicerPackages(parent_item)


def isApplicationFolder(folder):
    """
    Return True if folder an application folder.

    An application folder is expected to have the ``applicationPackageNameTemplate`` and
    ``extensionPackageNameTemplate`` metadata as well as parent folder named after
    :const:`constants.TOP_LEVEL_FOLDER_NAME`.
    """
    meta = folder.get('meta', {})
    if not all(k in meta for k in ('applicationPackageNameTemplate', 'extensionPackageNameTemplate')):
        return False

    parent_folder = Folder().load(folder['parentId'], force=True)
    if parent_folder['name'] != constants.TOP_LEVEL_FOLDER_NAME:
        return False

    return True


def isReleaseFolder(folder):
    """
    Return True if folder is a release folder.

    A release folder is expected to have the ``revision`` metadata as well as
    an application parent or grandparent folder (see :func:`isApplicationFolder`).
    """
    if 'revision' not in folder.get('meta', {}):
        return False

    parent_folder = Folder().load(folder['parentId'], force=True)
    grandparent_folder = Folder().load(parent_folder['parentId'], force=True)
    if not any([isApplicationFolder(parent_folder), isApplicationFolder(grandparent_folder)]):
        return False

    return True


def isDraftReleaseFolder(folder):
    """
    Return True if folder is a draft release folder.

    A draft release folder is expected to be a release folder (see :func:`isReleaseFolder`)
    and to have a parent folder named after :const:`constants.DRAFT_RELEASE_NAME`.
    """
    if not isReleaseFolder(folder):
        return False
    parent_folder = Folder().load(folder['parentId'], force=True)
    return parent_folder['name'] == constants.DRAFT_RELEASE_NAME


def getOrCreateReleaseFolder(application, user, app_revision):
    """
    Get or create the release folder associated with the application revision.

    :param application: The parent folder containing the release.
    :param user: The user to check access against or to create the new folder
    :param app_revision: The revision of the application.
    :return: The created/existing release folder.
    """
    release_folder = None
    # Find the release by metadata revision
    releases = Folder().childFolders(application, 'Folder', user=user)
    for folder in releases:
        revision = folder.get('meta', {}).get('revision', '')
        if revision == app_revision:
            release_folder = folder
            break
    if not release_folder:
        # Only the draft release in the list
        release_folder = list(Folder().childFolders(
            application,
            'Folder',
            user=user,
            filters={'name': constants.DRAFT_RELEASE_NAME}))
        if not release_folder:
            raise Exception('The %s folder not found.' % constants.DRAFT_RELEASE_NAME)
        release_folder = release_folder[0]

        revision_folder = list(Folder().childFolders(
            release_folder,
            'Folder',
            user=user,
            filters={'meta.revision': app_revision}))
        if revision_folder:
            revision_folder = revision_folder[0]
        else:
            revision_folder = Folder().createFolder(
                parent=release_folder,
                name=app_revision,
                parentType='Folder',
                public=release_folder['public'],
                creator=user)
            revision_folder = Folder().setMetadata(
                revision_folder,
                {'revision': app_revision})
        release_folder = revision_folder

    return release_folder


def getReleaseFolder(item, force=False):
    """
    Get item release folder.

    The release folder is either the one the with name matching
    :const:`constants.DRAFT_RELEASE_NAME` (e,g ``draft``) or the release
    folder (e.g ``1.0``).

    ::

        Applications
            |--- packages
            |        |----- Slicer
            |        |         |----- 1.0
            |        |         |        |---- Package1
            .        .         .        .
            |        |         |        |---- extensions
            |        |         |        |         |---- Extension1
            .        .         .        .         .
            .        .         .
            .        .         .
            |        |         |----- draft
            |        |         |        |--- r100
            |        |         |        |      |---- Package1
            .        .         .        .      .
            |        |         |        |      |----- extensions
            |        |         |        |      |          |---- Extension1

    :param item: A package or extension instance.
    :param force: If you explicitly want to circumvent access
                      checking on this resource, set this to True.
    :return: The parent release folder or None.
    """
    if not isSlicerPackages(item):
        return None

    folder = Folder().load(item['folderId'], level=AccessType.READ, force=force)

    # Check if item folder is "extensions". If yes, get parent folder.
    if folder['name'] == constants.EXTENSIONS_FOLDER_NAME:
        extensions_folder = folder
        folder = Folder().load(extensions_folder['parentId'], level=AccessType.READ, force=force)

    package_folder = folder

    # Check if grand parent folder is "draft". If yes return it.
    grantparent_folder = Folder().load(package_folder['parentId'], level=AccessType.READ, force=force)
    if grantparent_folder['name'] == constants.DRAFT_RELEASE_NAME:
        return grantparent_folder
    else:
        return package_folder


def deleteFolder(folder, progress, user):
    """Recursively delete a folder by ID.

    :param folder: The folder document to delete.
    :type folder: dict
    :param progress: A progress context to record progress on.
    :type progress: girder.utility.progress.ProgressContext or None.
    :param user: The user creating the progress.
    :type user: dict
    :return:
    """
    with ProgressContext(progress, user=user,
                         title='Deleting folder %s' % folder['name'],
                         message='Calculating folder size...') as ctx:
        # Don't do the subtree count if we weren't asked for progress
        if progress:
            ctx.update(total=Folder().subtreeCount(folder))
        Folder().remove(folder, progress=ctx)


def checkAccess(app_id, user):
    """
    Check user has access to the application.

    :param app_id: The ID of the application.
    :param user: The user to check access against.
    :raises girder.exceptions.AccessException: If the access check failed.
    :return:
    """
    Folder().load(app_id, level=AccessType.READ, user=user)
